# coding: UTF-8
import html
import os
import socket
import sys
import time
from pathlib import Path
from urllib.parse import parse_qs
from xml.etree import ElementTree

import requests
from requests.exceptions import Timeout
from requests.packages.urllib3.exceptions import TimeoutError, RequestError
from tqdm import tqdm

from nicotools import utils
from nicotools.utils import Msg, Err, URL, KeyGTI, KeyGetFlv

"""
使い方:

    nicodown --thumbnail --dest ".\Downloads" sm1 sm2 sm3 sm4 sm5
    nicodown --comment --video --thumbnail --dest ".\Downloads" sm1 sm2 sm3 sm4 sm5
    nicodown -cvt -d ".\Downloads" +ids.txt
    nicodown -cvt --xml -dest ".\Downloads" sm1

他のコマンド:
    引数がどの様に解釈されるかを確認したいとき (確認するだけで、プログラムは実行しません):
        nicodown.py --getthumbinfo sm12345678 --out ../file.txt --what

    ログ出力の詳細さを変える:
        nicodown --loglevel WARNING  # エラー以外表示しない
"""


def get_infos(queue, logger=None):
    """
    getthumbinfo APIから、細かな情報をもらってくる

    * comment_num       int
    * description       str
    * embeddable        int     # 0 or 1
    * file_name         str
    * first_retrieve    str     # 例えば: 2014-07-26 (もともとは: 2014-07-26T19:27:07+09:00)
    * last_res_body     str
    * length            str
    * length_seconds    int
    * mylist_counter    int
    * movie_type        str     # いずれか: mp4, flv, swf
    * no_live_play      int     # 0 or 1
    * size_high         int
    * size_low          int
    * tags              str
    * tags_list         list
    * thumbnail_url     str
    * title             str
    * user_id           int
    * user_name         str
    * user_icon_url     str
    * video_id          str
    * view_counter      int
    * v_or_t_id         str
    * watch_url         str

    :param list[str] | str queue: 動画IDのリスト
    :param NTLogger | None logger: ログ出力
    :rtype: dict[str, dict[str, int | str | list]]
    """
    if isinstance(queue, str):
        queue = [queue]
    message = Msg.nd_start_download.format(count=len(queue), ids=queue)
    if logger: logger.info(message)
    else: print(message)

    # データベースとして使うための辞書。削除や非公開の動画はここに入る
    lexikon = {}
    for video_id in utils.validator(queue):
        xmldata = requests.get(URL.URL_Info + video_id).text
        root = ElementTree.fromstring(xmldata)
        # 「status="ok"」 なら動画は生存 / 存在しない動画には「status="fail"」が返る
        if not root.get("status").lower() == "ok":
            if logger: logger.warning(Msg.nd_deleted_or_private.format(video_id))
            else: print(Msg.nd_deleted_or_private.format(video_id))
            continue
        else:
            # 各種情報を辞書に追加していく
            pocket = {}
            # document == "thumb" タグ
            document = root[0]  # type: ElementTree.Element
            finder = document.find

            # 日本語以外のタグが設定されている場合にそれらも巻き込む
            tag_list = [tagstr.text for tagstr in document.iter("tag")]
            # 「分:秒」形式を秒数に直すため分離する
            minute, second = finder(KeyGTI.LENGTH).text.split(":")

            pocket[KeyGTI.COMMENT_NUM]     = int(finder(KeyGTI.COMMENT_NUM).text)
            pocket[KeyGTI.DESCRIPTION]     = html.unescape(finder(KeyGTI.DESCRIPTION).text)
            pocket[KeyGTI.EMBEDDABLE]      = int(finder(KeyGTI.EMBEDDABLE).text)
            pocket[KeyGTI.FILE_NAME]       = utils.t2filename(finder(KeyGTI.TITLE).text)
            pocket[KeyGTI.FIRST_RETRIEVE]  = finder(KeyGTI.FIRST_RETRIEVE).text[:10]
            pocket[KeyGTI.LAST_RES_BODY]   = finder(KeyGTI.LAST_RES_BODY).text
            pocket[KeyGTI.LENGTH]          = "{}:{}".format(minute, second)
            pocket[KeyGTI.LENGTH_SECONDS]  = int(minute) * 60 + int(second)
            pocket[KeyGTI.MYLIST_COUNTER]  = int(finder(KeyGTI.MYLIST_COUNTER).text)
            pocket[KeyGTI.MOVIE_TYPE]      = finder(KeyGTI.MOVIE_TYPE).text.lower()
            pocket[KeyGTI.NO_LIVE_PLAY]    = int(finder(KeyGTI.NO_LIVE_PLAY).text)
            pocket[KeyGTI.SIZE_HIGH]       = int(finder(KeyGTI.SIZE_HIGH).text)
            pocket[KeyGTI.SIZE_LOW]        = int(finder(KeyGTI.SIZE_LOW).text)
            pocket[KeyGTI.TAGS]            = html.unescape(", ".join(tag_list))  # type: str
            pocket[KeyGTI.TAGS_LIST]       = tag_list  # type: list
            pocket[KeyGTI.THUMBNAIL_URL]   = finder(KeyGTI.THUMBNAIL_URL).text
            pocket[KeyGTI.TITLE]           = html.unescape(finder(KeyGTI.TITLE).text)
            pocket[KeyGTI.VIDEO_ID]        = video_id
            pocket[KeyGTI.VIEW_COUNTER]    = int(finder(KeyGTI.VIEW_COUNTER).text)
            pocket[KeyGTI.WATCH_URL]       = finder(KeyGTI.WATCH_URL).text
            pocket[KeyGTI.V_OR_T_ID]       = finder(KeyGTI.WATCH_URL).text.split("/")[-1]
            if video_id.startswith(("sm", "nm")):
                pocket[KeyGTI.USER_ID]         = int(finder(KeyGTI.USER_ID).text)
                pocket[KeyGTI.USER_NAME]       = html.unescape(finder(KeyGTI.USER_NAME).text)
                pocket[KeyGTI.USER_ICON_URL]   = finder(KeyGTI.USER_ICON_URL).text
            else:  # so1234 または 123456 の形式
                pocket[KeyGTI.CH_ID]           = int(finder(KeyGTI.CH_ID).text)
                pocket[KeyGTI.CH_NAME]         = html.unescape(finder(KeyGTI.CH_NAME).text)
                pocket[KeyGTI.CH_ICON_URL]     = finder(KeyGTI.CH_ICON_URL).text
            lexikon[video_id]                  = pocket
    return lexikon


class Video(utils.Canopy):
    def __init__(self, mail=None, password=None, logger=None, session=None):
        """
        動画をダウンロードする。

        :param str | None mail:
        :param str | None password:
        :param NTLogger logger:
        :param requests.Session session:
        """
        super().__init__(logger=logger)
        self.session = session or utils.LogIn(mail=mail, password=password).session

    def start(self, glossary, save_dir, chunk_size=1024*50):
        """

        :param dict[str, dict[str, int | str]] | list[str] glossary:
        :param str | Path save_dir:
        :param int chunk_size: 一度にサーバーに要求するファイルサイズ
        :rtype: bool
        """
        utils.check_arg(locals())
        self.logger.debug("Directory to save in: {}".format(save_dir))
        self.logger.debug("Dictionary of Videos: {}".format(glossary))
        self.save_dir = utils.make_dir(save_dir)
        if isinstance(glossary, list):
            glossary = get_infos(glossary, self.logger)
        self.glossary = glossary
        self.logger.info(Msg.nd_start_dl_video.format(
            count=len(self.glossary), ids=list(self.glossary)))

        for index, video_id in enumerate(self.glossary.keys()):
            self.logger.info(
                Msg.nd_download_video.format(
                    index + 1, len(glossary), video_id,
                    self.glossary[video_id][KeyGTI.TITLE]))
            self._download(video_id, chunk_size)
            if len(glossary) > 1:
                time.sleep(1)
        return True

    def _download(self, video_id, chunk_size=1024 * 50):
        """
        :param str video_id: 動画ID (e.g. sm1234)
        :param int chunk_size: 一度にサーバーに要求するファイルサイズ
        :rtype: bool
        """
        utils.check_arg(locals())
        db = self.glossary[video_id]
        if video_id.startswith("so"):
            redirected = self.session.get(URL.URL_Watch + video_id).url.split("/")[-1]
            db[KeyGTI.V_OR_T_ID] = redirected
        self.logger.debug("Video ID and its Thread ID (of officials):"
                          " {}".format(video_id, db[KeyGTI.V_OR_T_ID]))

        response = utils.get_from_getflv(
            db[KeyGTI.V_OR_T_ID], self.session, self.logger)

        vid_url = response[KeyGetFlv.VIDEO_URL]
        is_premium = response[KeyGetFlv.IS_PREMIUM]

        # 動画視聴ページに行ってCookieをもらってくる
        self.session.head(URL.URL_Watch + video_id)
        # connect timeoutを10秒, read timeoutを30秒に設定
        # ↓この時点ではダウンロードは始まらず、ヘッダーだけが来ている
        video_data = self.session.get(url=vid_url, stream=True, timeout=(10.0, 30.0))
        db[KeyGTI.FILE_SIZE] = int(video_data.headers["content-length"])
        self.logger.debug("File Size: {} (Premium: {})".format(
            db[KeyGTI.FILE_SIZE], [False, True][int(is_premium)]))

        return self._saver(video_id, video_data, chunk_size)

    def _saver(self, video_id, video_data, chunk_size):
        """

        :param str video_id:
        :param requests.Response video_data: 動画ファイルのURL
        :param int chunk_size: 一度にサーバーに要求するファイルサイズ
        :rtype: bool
        """
        file_path = self.make_name(video_id, self.glossary[video_id][KeyGTI.MOVIE_TYPE])
        self.logger.debug("File Path: {}".format(file_path))
        db = self.glossary[video_id]

        if tqdm is None:
            with file_path.open("wb") as f:
                [f.write(chunk) for chunk in
                 video_data.iter_content(chunk_size=chunk_size) if chunk]
        else:
            with tqdm(total=db[KeyGTI.FILE_SIZE], leave=False,
                      unit="B", unit_scale=True, file=sys.stdout) as pbar:
                with file_path.open("wb") as f:
                    for chunk in video_data.iter_content(chunk_size=chunk_size):
                        if chunk:
                            pbar.update(f.write(chunk))
        self.logger.info(Msg.nd_download_done.format(path=file_path))
        return True


class Thumbnail(utils.Canopy):
    def __init__(self, logger=None):
        """
        :param NTLogger logger:
        """
        super().__init__(logger=logger)

    def start(self, glossary, save_dir, is_large=True):
        """

        :param dict[str, dict[str, int | str]] | list[str] glossary:
        :param str | Path save_dir:
        :param bool is_large: 大きいサムネイルを取りに行くかどうか
        :rtype: bool
        """
        utils.check_arg(locals())
        self.logger.debug("Directory to save in: {}".format(save_dir))
        self.logger.debug("Dictionary of Videos: {}".format(glossary))
        if isinstance(glossary, list):
            glossary = get_infos(glossary, self.logger)
        self.glossary = glossary
        self.save_dir = utils.make_dir(save_dir)

        self.logger.info(Msg.nd_start_dl_pict.format(
            count=len(self.glossary), ids=list(self.glossary)))

        for index, video_id in enumerate(self.glossary.keys()):
            self.logger.info(
                Msg.nd_download_pict.format(
                    index + 1, len(glossary), video_id,
                    self.glossary[video_id][KeyGTI.TITLE]))
            self._download(video_id, is_large)
        return True

    def _download(self, video_id, is_large=True):
        """
        :param str video_id: 動画ID (e.g. sm1234)
        :param bool is_large: 大きいサムネイルを取りに行くかどうか
        :rtype: bool
        """
        utils.check_arg(locals())
        url = self.glossary[video_id][KeyGTI.THUMBNAIL_URL]
        if is_large:
            url += ".L"
        image_data = self._worker(video_id, url, is_large)
        if not image_data:
            return False
        return self._saver(video_id, image_data)

    def _worker(self, video_id, url, is_large=True):
        """
        サムネイル画像をダウンロードしにいく。

        :param str video_id: 動画ID (e.g. sm1234)
        :param str url: 画像のURL
        :param bool is_large: 大きいサムネイルを取りに行くかどうか
        :rtype: bool | requests.Response
        """
        utils.check_arg(locals())
        db = self.glossary[video_id]
        with requests.Session() as session:
            try:
                # connect timeoutを5秒, read timeoutを10秒に設定
                response = session.get(url=url, timeout=(5.0, 10.0))
                if response.ok:
                    return response
                # 大きいサムネイルを求めて404が返ってきたら標準の大きさで試す
                if response.status_code == 404:
                    if is_large and url.endswith(".L"):
                        return self._worker(video_id, url[:-2], is_large=False)
                    else:
                        self.logger.error(Err.connection_404.format(
                            video_id, db[KeyGTI.TITLE]))
                        return False
            except (TypeError, ConnectionError,
                    socket.timeout, Timeout, TimeoutError, RequestError) as e:
                self.logger.debug("An exception occurred: {}".format(e))
                if is_large and url.endswith(".L"):
                    return self._worker(video_id, url[:-2], is_large=False)
                else:
                    self.logger.error(Err.connection_timeout.format(video_id)
                                      + " (タイトル: {})".format(db[KeyGTI.TITLE]))
                    return False

    def _saver(self, video_id, image_data, _=None):
        """

        :param str video_id: 動画ID (e.g. sm1234)
        :param requests.Response image_data: 画像のデータ
        :return:
        """
        file_path = self.make_name(video_id, "jpg")
        self.logger.debug("File Path: {}".format(file_path))

        with file_path.open('wb') as f:
            f.write(image_data.content)
        self.logger.info(Msg.nd_download_done.format(path=file_path))
        return True


class Comment(utils.Canopy):
    def __init__(self, mail=None, password=None, logger=None, session=None):
        """
        :param str | None mail:
        :param str | None password:
        :param NTLogger logger:
        :param requests.Session session:
        """
        super().__init__(logger=logger)
        self.session = session or utils.LogIn(mail=mail, password=password).session

    def start(self, glossary, save_dir, xml=False):
        """

        :param dict[str, dict[str, int | str]] | list[str] glossary:
        :param str | Path save_dir:
        :param bool xml:
        """
        utils.check_arg(locals())
        self.logger.debug("Directory to save in: {}".format(save_dir))
        self.logger.debug("Dictionary of Videos: {}".format(glossary))
        self.logger.debug("Download XML? : {}".format(xml))
        if isinstance(glossary, list):
            glossary = get_infos(glossary, self.logger)
        self.glossary = glossary
        self.save_dir = utils.make_dir(save_dir)
        self.logger.info(Msg.nd_start_dl_comment.format(
            count=len(self.glossary), ids=list(self.glossary)))
        for index, video_id in enumerate(self.glossary.keys()):
            self.logger.info(
                Msg.nd_download_comment.format(
                    index + 1, len(glossary), video_id,
                    self.glossary[video_id][KeyGTI.TITLE]))
            self._download(video_id, xml)
            if len(self.glossary) > 1:
                time.sleep(1.5)
        return True

    def _download(self, video_id, xml=False):
        """
        :param str video_id: 動画ID (e.g. sm1234)
        :param bool xml:
        :rtype: bool
        """
        utils.check_arg(locals())
        db = self.glossary[video_id]
        if video_id.startswith("so"):
            redirected = self.session.get(URL.URL_Watch + video_id).url.split("/")[-1]
            db[KeyGTI.V_OR_T_ID] = redirected
        self.logger.debug("Video ID and its Thread ID (of officials):"
                          " {}".format(video_id, db[KeyGTI.V_OR_T_ID]))

        response = utils.get_from_getflv(
            db[KeyGTI.V_OR_T_ID], self.session, self.logger)

        if response is None:
            time.sleep(4)
            print(Err.waiting_for_permission)
            time.sleep(4)
            return self._download(video_id, xml)

        thread_id = response[KeyGetFlv.THREAD_ID]
        msg_server = response[KeyGetFlv.MSG_SERVER]
        user_id = response[KeyGetFlv.USER_ID]
        user_key = response[KeyGetFlv.USER_KEY]

        opt_thread_id = response[KeyGetFlv.OPT_THREAD_ID]
        needs_key = response[KeyGetFlv.NEEDS_KEY]

        if xml and video_id.startswith(("sm", "nm")):
            req_param = self.make_param_xml(thread_id, user_id)
            self.logger.debug("Posting Parameters: {}".format(req_param))

            res_com = self.session.post(url=msg_server, data=req_param)
            comment_data = res_com.text.replace("><", ">\n<")
        else:
            if video_id.startswith(("sm", "nm")):
                req_param = self.make_param_json(
                    False, user_id, user_key, thread_id)
            else:
                thread_key, force_184 = self.get_thread_key(db[KeyGTI.V_OR_T_ID],
                                                            needs_key)
                req_param = self.make_param_json(
                    True, user_id, user_key, thread_id,
                    opt_thread_id, thread_key, force_184)

            self.logger.debug("Posting Parameters: {}".format(req_param))
            res_com = self.session.post(
                url=URL.URL_Msg_JSON,
                json=req_param)
            comment_data = res_com.text.replace("}, ", "},\n")

        comment_data = comment_data.encode(res_com.encoding).decode("utf-8")
        return self._saver(video_id, comment_data, xml)

    def _saver(self, video_id, comment_data, xml):
        """

        :param str video_id:
        :param str comment_data:
        :param bool xml:
        :return:
        """
        utils.check_arg(locals())
        if xml and video_id.startswith(("sm", "nm")):
            extention = "xml"
        else:
            extention = "json"

        file_path = self.make_name(video_id, extention)
        self.logger.debug("File Path: {}".format(file_path))
        with file_path.open("w", encoding="utf-8") as f:
            f.write(comment_data + "\n")
        self.logger.info(Msg.nd_download_done.format(path=file_path))
        return True

    def get_thread_key(self, thread_id, needs_key):
        """
        専用のAPIにアクセスして thread_key を取得する。

        :param str thread_id:
        :param str needs_key:
        :rtype: tuple[str, str]
        """
        utils.check_arg(locals())
        if not needs_key == "1":
            self.logger.debug("Video ID (or Thread ID): {},"
                              " needs_key: {}".format(thread_id, needs_key))
            return "", "0"
        response = self.session.get(URL.URL_GetThreadKey, params={"thread": thread_id})
        self.logger.debug("Response from GetThreadKey API: {}".format(response.text))
        parameters = parse_qs(response.text)
        threadkey = parameters["threadkey"][0]  # type: str
        force_184 = parameters["force_184"][0]  # type: str
        return threadkey, force_184

    def make_param_xml(self, thread_id, user_id):
        """
        コメント取得用のxmlを構成する。

        fork="1" があると投稿者コメントを取得する。
        0-99999:9999,1000: 「0分～99999分までの範囲で
        一分間あたり9999件、直近の1000件を取得する」の意味。

        :param str thread_id:
        :param str user_id:
        :rtype: str
        """
        utils.check_arg(locals())
        self.logger.debug("Arguments: {}".format(locals()))
        return '<packet>' \
              '<thread thread="{0}" user_id="{1}" version="20090904" scores="1"/>' \
              '<thread thread="{0}" user_id="{1}" version="20090904" scores="1"' \
              ' fork="1" res_from="-1000"/>' \
              '<thread_leaves thread="{0}" user_id="{1}" scores="1">' \
              '0-99999:9999,1000</thread_leaves>' \
              '</packet>'.format(thread_id, user_id)

    def make_param_json(self, official_video, user_id, user_key, thread_id,
                        optional_thread_id=None, thread_key=None, force_184=None):
        """
        コメント取得用のjsonを構成する。

        fork="1" があると投稿者コメントを取得する。
        0-99999:9999,1000: 「0分～99999分までの範囲で
        一分間あたり9999件、直近の1000件を取得する」の意味。

        :param bool official_video: 公式動画なら True
        :param str user_id:
        :param str user_key:
        :param str thread_id:
        :param str | None optional_thread_id:
        :param str | None thread_key:
        :param str | None force_184:
        :rtype: list[dict]
        """
        utils.check_arg({"official_video": official_video, "user_id": user_id,
                         "user_key": user_key, "thread_id": thread_id})
        self.logger.debug("Arguments of creating JSON: {}".format(locals()))
        result = [
            {"ping": {"content": "rs:0"}},
            {"ping": {"content": "ps:0"}},
            {
                "thread": {
                    "thread"     : optional_thread_id or thread_id,
                    "version"    : "20090904",
                    "language"   : 0,
                    "user_id"    : user_id,
                    "with_global": 1,
                    "scores"     : 1,
                    "nicoru"     : 0,
                    "userkey"    : user_key
                }
            },
            {"ping": {"content": "pf:0"}},
            {"ping": {"content": "ps:1"}},
            {
                "thread_leaves": {
                    "thread"  : optional_thread_id or thread_id,
                    "language": 0,
                    "user_id" : user_id,
                    # "content" : "0-4:100,250",  # 公式仕様のデフォルト値
                    "content" : "0-99999:9999,1000",
                    "scores"  : 1,
                    "nicoru"  : 0,
                    "userkey" : user_key
                }
            },
            {"ping": {"content": "pf:1"}}
        ]

        if official_video:
            result += [{"ping": {"content": "ps:2"}},
                {
                    "thread": {
                        "thread"     : thread_id,
                        "version"    : "20090904",
                        "language"   : 0,
                        "user_id"    : user_id,
                        "force_184"  : force_184,
                        "with_global": 1,
                        "scores"     : 1,
                        "nicoru"     : 0,
                        "threadkey"  : thread_key
                    }
                },
                {"ping": {"content": "pf:2"}},
                {"ping": {"content": "ps:3"}},
                {
                    "thread_leaves": {
                        "thread"   : thread_id,
                        "language" : 0,
                        "user_id"  : user_id,
                        # "content"  : "0-4:100,250",  # 公式仕様のデフォルト値
                        "content"  : "0-99999:9999,1000",
                        "scores"   : 1,
                        "nicoru"   : 0,
                        "force_184": force_184,
                        "threadkey": thread_key
                    }
                },
                {"ping": {"content": "pf:3"}}]
        result += [{"ping": {"content": "rf:0"}}]
        return result


def main(args):
    """
    メイン。

    :param args: ArgumentParser.parse_args() によって解釈された引数
    :rtype: bool
    """
    is_debug = int(os.getenv("PYTHON_TEST", 0))
    mailadrs = args.mail[0] if args.mail else None
    password = args.password[0] if args.password else None

    """ エラーの除外 """
    if hasattr(args, "dmc"):
        sys.exit(Err.unexpected_commands.format("--dmc"))
    if hasattr(args, "smile"):
        sys.exit(Err.unexpected_commands.format("--smile"))
    videoid = utils.validator(args.VIDEO_ID)
    if not videoid:
        sys.exit(Err.invalid_videoid)
    if not (args.getthumbinfo or args.thumbnail or args.comment or args.video):
        sys.exit(Err.not_specified.format("--thumbnail、 --comment、 --video のいずれか"))

    if args.getthumbinfo:
        file_name = args.out[0] if isinstance(args.out, list) else None
        return utils.print_info(videoid, file_name)

    """ 本筋 """
    log_level = "DEBUG" if is_debug else args.loglevel
    logger = utils.NTLogger(log_level=log_level, file_name=utils.LOG_FILE_ND)
    destination = utils.make_dir(args.dest[0])
    database = get_infos(videoid, logger=logger)

    res_t = False
    if args.thumbnail:
        res_t = Thumbnail(logger=logger).start(database, destination)
        if not (args.comment or args.video):
            # サムネイルのダウンロードだけならここで終える。
            return res_t

    session = utils.LogIn(mail=mailadrs, password=password).session

    res_c = False
    if args.comment:
        res_c = Comment(logger=logger, session=session).start(database, destination, args.xml)

    res_v = False
    if args.video:
        res_v = Video(logger=logger, session=session).start(database, destination)

    return res_c | res_v | res_t



# coding: UTF-8
import asyncio
import functools
import html
import json
import os
import re
import sys
from pathlib import Path
from string import Template
from typing import Dict, Union, Optional, List
from urllib.parse import parse_qs, unquote

import aiohttp
from bs4 import BeautifulSoup, Tag
from tqdm import tqdm

from nicotools import utils
from nicotools.utils import Msg, Err, URL, KeyGetFlv, KeyGTI, KeyDmc, NotLoginError


class Info(utils.CanopyAsync):
    def __init__(self,
                 mail: str=None, password: str=None,
                 logger: utils.NTLogger=None,
                 session: aiohttp.ClientSession=None,
                 limit: int=4,
                 sieve=True,
                 loop: asyncio.AbstractEventLoop=None,
                 return_session=False,
                 interval: Union[int, float]=5,
                 backoff: Union[int, float]=3,
                 retries: Union[int, float]=3,
                 ):
        """
        動画視聴ページから様々なデータを集める。

        :param Optional[str] mail: メールアドレス
        :param Optional[str] password: パスワード
        :param T <= logging.logger logger: ロガーのインスタンス
        :param aiohttp.ClientSession session:
        :param bool return_session: データを取り終わった後にセッションを返すか閉じるか
        :param int limit: 同時にアクセスする最大数
        :param bool sieve: 非公開や削除済み動画を除外するかどうか
        :param asyncio.AbstractEventLoop loop: イベントループ
        :param Union[int, float] interval: うまくいかなかった場合の待ち時間
        :param Union[int, float] backoff: 待ち時間の増大倍率
        :param Union[int,float] retries: 再試行回数
        """
        super().__init__(loop=loop, logger=logger)
        self.__mail = mail
        self.__password = password
        self.session = session or self.loop.run_until_complete(self.get_session())
        self.__parallel_limit = limit
        self.__do_sieve = sieve
        self.__return_session = return_session
        self.interval = interval
        self.backoff = backoff
        self.retries = retries

    async def get_session(self) -> aiohttp.ClientSession:
        """
        aiohttp のセッションを返す。

        :rtype: aiohttp.ClientSession
        """
        if self.session:
            return self.session
        else:
            login = utils.LogIn(mail=self.__mail, password=self.__password)
            if login.is_login:
                cook = login.cookie
            else:
                login.get_session(utils.LogIn.ask_credentials())
                cook = login.cookie
            self.logger.debug("Object ID of cookie (Info): {}".format(id(cook)))
            return aiohttp.ClientSession(cookies=cook)

    def get_data(self, video_ids: List) -> Dict:
        """
        動画やコメントのダウンロードに必要なデータを集めてくる。

        :param List video_ids:
        :rtype: Dict
        """
        glossary = utils.validator(video_ids)

        self.logger.info(Msg.nd_start_download.format(
            count=len(glossary), ids=glossary))

        infos = self.loop.run_until_complete(
            asyncio.gather(*[self._retrieve_info(_id) for _id in video_ids]))
        result = {_id: _info for _id, _info in zip(video_ids, infos)}
        if not self.__return_session:
            self.session.close()
        sieved_result = self.siever(result)
        return sieved_result

    def siever(self, infos: Dict) -> Dict:
        """
        非公開や削除済み動画をふるいにかける。

        :param Dict infos:
        :rtype: Dict
        """
        good = {_id: info for _id, info in infos.items()
                  if info[KeyDmc.IS_PUBLIC] and not info[KeyDmc.IS_DELETED]}
        bad = list(set(infos) - set(good))
        if len(bad) > 0:
            self.logger.info(Msg.nd_deleted_or_private.format(bad))
        if self.__do_sieve:
            return good
        else:
            return infos

    async def _retrieve_info(self, video_id: str) -> Dict:
        interval = self.interval
        backoff = self.backoff
        attempt = max(0, self.retries) + 1
        url = URL.URL_Watch + video_id
        self.logger.debug("_worker: {}".format(locals()))

        async with asyncio.Semaphore(self.__parallel_limit):
            st = 0
            while attempt > 0:
                attempt -= 1
                async with self.session.get(url) as response:  # type: aiohttp.ClientResponse
                    self.logger.debug("Video ID: {}, Status Code: {}".format(
                        video_id, response.status))
                    if response.status == 200:
                        info_data = await response.text()
                        return self._junction(info_data)
                    # ステータスコードが400番台なら例外を出す
                    elif 400 <= response.status < 500:
                        response.raise_for_status()
                    elif 500 <= response.status < 600:
                        await asyncio.sleep(interval/2)
                        print(Err.waiting_for_permission)
                        await asyncio.sleep(interval/2)
                        interval *= backoff
                        st = response.status
                    else:
                        st = response.status
                        break
            raise aiohttp.errors.HttpProcessingError(
                code=st, message=Err.connection_timeout.format(video_id))

    def _pick_info_from_watch_api(self, content: str) -> \
            Dict[str, Union[str, int, List[str], bool]]:
        """
        watchAPIDataContainer を含む HTML から情報を取り出す。

        :param str content: HTMLの文字列
        :rtype: Dict
        """
        watch_api = json.loads(content)
        flash_vars = watch_api["flashvars"]
        flvinfo = utils.extract_getflv(unquote(flash_vars["flvInfo"]))
        if "dmcInfo" in flash_vars:
            dmc_info = json.loads(unquote(flash_vars["dmcInfo"]))
            session_api = dmc_info["session_api"]
        else:
            dmc_info = None
            session_api = None

        info = {
            KeyDmc.VIDEO_ID     : flash_vars["videoId"],  # str
            KeyDmc.VIDEO_URL_SM : flvinfo[KeyGetFlv.VIDEO_URL],  # str
            KeyDmc.TITLE        : flash_vars["videoTitle"],  # str
            KeyDmc.FILE_NAME    : utils.t2filename(flash_vars["videoTitle"]),
            KeyDmc.FILE_SIZE    : None,  # int
            KeyDmc.THUMBNAIL_URL: flash_vars["thumbImage"],  # str
            KeyDmc.ECO          : flash_vars.get("eco") or 0,  # int
            KeyDmc.MOVIE_TYPE   : flash_vars["movie_type"],  # str
            # KeyDmc.IS_DMC       : int(flash_vars["isDmc"]),  # int
            KeyDmc.DELETED      : int(flash_vars["deleted"]),  # int
            KeyDmc.IS_DELETED   : watch_api["videoDetail"]["isDeleted"],  # bool
            KeyDmc.IS_PUBLIC    : watch_api["videoDetail"]["is_public"],  # bool
            KeyDmc.IS_OFFICIAL  : watch_api["videoDetail"]["is_official"],  # bool
            KeyDmc.IS_PREMIUM   : watch_api["viewerInfo"]["isPremium"],  # bool

            # この6つはコメントのダウンロードに必要
            KeyDmc.USER_ID      : flvinfo[KeyGetFlv.USER_ID],  # int
            KeyDmc.USER_KEY     : flvinfo[KeyGetFlv.USER_KEY],  # str
            KeyDmc.MSG_SERVER   : flvinfo[KeyGetFlv.MSG_SERVER],  # str
            KeyDmc.THREAD_ID    : flvinfo[KeyGetFlv.THREAD_ID],  # int
            KeyDmc.OPT_THREAD_ID: flvinfo[KeyGetFlv.OPT_THREAD_ID],  # int なければ None
            KeyDmc.NEEDS_KEY    : flvinfo[KeyGetFlv.NEEDS_KEY],  # int なければ None

            # DMCサーバーからの動画のダウンロードに必要
            KeyDmc.API_URL      : None,
            KeyDmc.RECIPE_ID    : None,
            KeyDmc.CONTENT_ID   : None,
            KeyDmc.VIDEO_SRC_IDS: None,
            KeyDmc.AUDIO_SRC_IDS: None,
            KeyDmc.HEARTBEAT    : None,
            KeyDmc.TOKEN        : None,
            KeyDmc.SIGNATURE    : None,
            KeyDmc.AUTH_TYPE    : None,
            KeyDmc.C_K_TIMEOUT  : None,
            KeyDmc.SVC_USER_ID  : None,
            KeyDmc.PLAYER_ID    : None,
            KeyDmc.PRIORITY     : None,
        }
        if dmc_info:
            info.update({
                KeyDmc.API_URL      : session_api["api_urls"][0],  # str
                KeyDmc.RECIPE_ID    : session_api["recipe_id"],  # str
                KeyDmc.CONTENT_ID   : session_api["content_id"],  # str
                KeyDmc.VIDEO_SRC_IDS: session_api["videos"],  # List[str]
                KeyDmc.AUDIO_SRC_IDS: session_api["audios"],  # List[str]
                KeyDmc.HEARTBEAT    : session_api["heartbeat_lifetime"],  # int
                KeyDmc.TOKEN        : session_api["token"],  # str
                KeyDmc.SIGNATURE    : session_api["signature"],  # str
                KeyDmc.AUTH_TYPE    : session_api["auth_types"]["http"],  # str
                KeyDmc.C_K_TIMEOUT  : session_api["content_key_timeout"],  # int
                KeyDmc.SVC_USER_ID  : info[KeyDmc.USER_ID],
                KeyDmc.PLAYER_ID    : session_api["player_id"],  # str
                KeyDmc.PRIORITY     : session_api["priority"],  # int
                # KeyDmc.OPT_THREAD_ID: dmc_info["thread"]["optional_thread_id"],  # int
                # KeyDmc.NEEDS_KEY    : dmc_info["thread"]["thread_key_required"],  # bool
            })
        return info

    def _pick_info_from_data_api(self, content: str) -> \
            Dict[str, Union[str, int, List[str], bool]]:  # pragma: no cover
        """
        data-api-data 属性を持つタグがあるHTMLから情報を取り出す。

        :param str content: HTMLの文字列
        :rtype: Dict
        """
        j = json.loads(content)
        _video = j["video"]
        if "dmcInfo" in _video:
            dmc_info = _video["dmcInfo"]
            session_api = dmc_info["session_api"]
        else:
            dmc_info = None
            session_api = None

        info = {
            KeyDmc.VIDEO_ID     : _video["id"],  # type: str
            KeyDmc.VIDEO_URL_SM : _video["source"],  # type: str
            KeyDmc.TITLE        : _video["originalTitle"],  # type: str
            KeyDmc.FILE_NAME    : utils.t2filename(_video["originalTitle"]),
            KeyDmc.FILE_SIZE    : None,
            KeyDmc.THUMBNAIL_URL: _video["thumbnailURL"],  # type: str
            KeyDmc.ECO          : j["context"]["isEconomy"],  # type: int
            KeyDmc.MOVIE_TYPE   : _video["movieType"],  # type: str
            KeyDmc.DELETED      : dmc_info["video"]["deleted"],  # type: int
            KeyDmc.IS_DELETED   : _video["isDeleted"],  # type: bool
            KeyDmc.IS_PUBLIC    : _video["isPublic"],  # type: bool
            KeyDmc.IS_OFFICIAL  : _video["isOfficial"],  # type: bool
            KeyDmc.IS_PREMIUM   : j["viewer"]["isPremium"],  # type: bool
            # KeyDmc.IS_DMC       : None,

            # この6つはコメントのダウンロードに必要
            KeyDmc.USER_ID      : j["viewer"]["id"],  # type: int
            KeyDmc.USER_KEY     : j["context"]["userkey"],  # type: str
            KeyDmc.OPT_THREAD_ID: None,  # int なければ None
            KeyDmc.NEEDS_KEY    : None,  # int なければ None
            # ただしこの2つは dmcInfo にしかない。
            # watchAPI版との整合性のために初期化しておく。
            KeyDmc.MSG_SERVER   : None,
            KeyDmc.THREAD_ID    : None,

            # DMCサーバーからの動画のダウンロードに必要
            KeyDmc.API_URL      : None,
            KeyDmc.RECIPE_ID    : None,
            KeyDmc.CONTENT_ID   : None,
            KeyDmc.VIDEO_SRC_IDS: None,
            KeyDmc.AUDIO_SRC_IDS: None,
            KeyDmc.HEARTBEAT    : None,
            KeyDmc.TOKEN        : None,
            KeyDmc.SIGNATURE    : None,
            KeyDmc.AUTH_TYPE    : None,
            KeyDmc.C_K_TIMEOUT  : None,
            KeyDmc.SVC_USER_ID  : None,
            KeyDmc.PLAYER_ID    : None,
            KeyDmc.PRIORITY     : None,
        }

        if dmc_info:
            info.update({
                KeyDmc.MSG_SERVER   : dmc_info["thread"]["server_url"],  # type: str
                KeyDmc.THREAD_ID    : dmc_info["thread"]["thread_id"],  # type: int
                KeyDmc.API_URL      : session_api["api_urls"][0],  # type: str
                KeyDmc.RECIPE_ID    : session_api["recipe_id"],  # type: str
                KeyDmc.CONTENT_ID   : session_api["content_id"],  # type: str
                KeyDmc.VIDEO_SRC_IDS: session_api["videos"],  # type: List[str]
                KeyDmc.AUDIO_SRC_IDS: session_api["audios"],  # type: List[str]
                KeyDmc.HEARTBEAT    : session_api["heartbeat_lifetime"],  # type: int
                KeyDmc.TOKEN        : session_api["token"],  # type: str
                KeyDmc.SIGNATURE    : session_api["signature"],  # type: str
                KeyDmc.AUTH_TYPE    : session_api["auth_types"]["http"],  # type: str
                KeyDmc.C_K_TIMEOUT  : session_api["content_key_timeout"],  # type: int
                KeyDmc.SVC_USER_ID  : info[KeyDmc.USER_ID],
                KeyDmc.PLAYER_ID    : session_api["player_id"],  # type: str
                KeyDmc.PRIORITY     : session_api["priority"],  # type: int
                KeyDmc.OPT_THREAD_ID: dmc_info["thread"]["optional_thread_id"],  # type: int
                KeyDmc.NEEDS_KEY    : int(dmc_info["thread"]["thread_key_required"]),  # type: int
            })
        return info

    def _junction(self, content: str) -> Dict[str, Union[str, int, List[str], bool]]:
        """
        動画視聴ページのHTMLから必要な情報を取り出す。

        HTML構造は複数あり、ものによって内容が異なる。適切な担当者へ振り向ける。

        :param str content:
        :rtype: Dict
        """
        soup = BeautifulSoup(content, "html.parser")
        _not_login = soup.select("#Login_nico")
        if _not_login:
            raise NotLoginError("ログインしてないよ")
        _data_api = soup.select("#js-initial-watch-data")
        _watch_api = soup.select("#watchAPIDataContainer")
        if _data_api:
            return self._pick_info_from_data_api(_data_api[0]["data-api-data"])
        elif _watch_api:
            return self._pick_info_from_watch_api(_watch_api[0].text)
        else:
            file_name = "_#_niconico_#_.html"
            with open(file_name, "w", encoding="utf-8") as fd:
                fd.write(content)
            raise AttributeError("Unknown HTML. For debug purpose,"
                                 " the content has been saved in {}."
                                 " You can safely delete it.".format(file_name))


class Thumbnail(utils.CanopyAsync):
    def __init__(self,
                 logger: utils.NTLogger=None,
                 limit: int=8,
                 session: aiohttp.ClientSession=None,
                 return_session: bool=False,
                 loop: asyncio.AbstractEventLoop=None,
                 ):
        """
        サムネイル画像をダウンロードする。

        :param T<= logging.logger logger: ロガー
        :param int limit: 同時にアクセスする最大数
        :param aiohttp.ClientSession session:
        :param bool return_session: セッションを返すか閉じるか
        :param asyncio.AbstractEventLoop loop: イベントループ
        """
        super().__init__(loop=loop, logger=logger)
        self.undone = []
        self.__bucket = {}
        self.session = session or self.loop.run_until_complete(self.get_session())
        self.__return_session = return_session
        self.__parallel_limit = limit

    async def get_session(self) -> aiohttp.ClientSession:
        return aiohttp.ClientSession()

    def close(self):
        self.session.close()

    def start(self, glossary: Union[List, Dict],
              save_dir: Union[str, Path], is_large=True):
        """
        ダウンロードを開始する。

        :param dict[str, dict[str, int | str]] | list[str] glossary:
         動画の情報が入った辞書またはIDのリスト
        :param str | Path save_dir:
        :param bool is_large: 大きいサムネイルを取りに行くかどうか
        :rtype: bool | aiohttp.ClientSession
        """
        utils.check_arg(locals())

        if isinstance(glossary, list):
            glossary = utils.validator(glossary)
            glossary = self.loop.run_until_complete(self._get_infos(glossary))
        self.glossary = glossary
        self.logger.debug("Dictionary of Videos: {}".format(self.glossary))

        self.save_dir = utils.make_dir(save_dir)
        self.logger.debug("Directory to save in: {}".format(self.save_dir))

        if len(self.glossary) > 0:
            self.logger.info(Msg.nd_start_dl_pict.format(
                count=len(self.glossary), ids=list(self.glossary)))

            self.loop.run_until_complete(self._download(list(self.glossary)))
            while len(self.undone) > 0:
                self.logger.debug("いくつか残ってる。{}".format(self.undone))
                self.loop.run_until_complete(self._download(self.undone, False))
            self.logger.debug("全部終わった")
        if not self.__return_session:
            self.session.close()
        return self

    async def _download(self, video_ids: list, islarge: bool=True) -> None:
        urls = self._make_urls(video_ids, islarge)

        futures = []
        for idx, video_id, _url in zip(range(len(video_ids)), video_ids, urls):
            coro = self._worker(idx, video_id, _url)
            f = asyncio.ensure_future(coro)
            f.add_done_callback(functools.partial(self._saver, video_id))
            futures.append(f)
        await asyncio.wait(futures, loop=self.loop)

    async def _worker(self, idx: int, video_id: str, url: str) -> Optional[bytes]:
        async with asyncio.Semaphore(self.__parallel_limit):

            self.logger.info(
                Msg.nd_download_pict.format(
                    idx + 1, len(self.glossary), video_id,
                    self.glossary[video_id][KeyGTI.TITLE]))

            try:
                async with self.session.get(url, timeout=10) as response:
                    self.logger.debug("Video ID: {}, Status Code: {}".format(video_id, response.status))
                    if response.status == 200:
                        # ダウンロードに成功したら「未完了のリスト」から取り除く
                        if video_id in self.undone:
                            self.undone.remove(video_id)
                        return await response.content.read()
                    else:
                        self.undone.append(video_id)
                        return None
            except asyncio.TimeoutError:
                self.logger.warning("{} が時間切れ".format(video_id))
                self.undone.append(video_id)
                return None

    def _saver(self, video_id: str, coroutine: asyncio.Task) -> None:
        image_data = coroutine.result()
        if image_data:
            file_path = self.make_name(video_id, "jpg")
            self.logger.debug("File Path: {}".format(file_path))

            with file_path.open('wb') as f:
                f.write(image_data)
            self.logger.info(Msg.nd_download_done.format(path=file_path))

    def _make_urls(self, video_ids: List[str], is_large: bool=True) -> List[str]:
        if is_large:
            urls = ["{0}.L".format(self.glossary[_id][KeyGTI.THUMBNAIL_URL]) for _id in video_ids]
        else:
            urls = [self.glossary[_id][KeyGTI.THUMBNAIL_URL] for _id in video_ids]
        return urls

    async def _get_infos(self, queue: List[str]) -> Dict[str, Dict]:
        """
        getthumbinfo APIから、細かな情報をもらってくる

        * file_name         str
        * thumbnail_url     str
        * title             str
        * video_id          str

        :param list[str] queue: 動画IDのリスト
        :rtype: dict[str, dict]
        """
        tasks = [self._get_infos_worker(video_id) for video_id in queue]
        await asyncio.wait(tasks, loop=self.loop)
        bad = list(set(queue) - set(self.__bucket))
        if len(bad) > 0:
            self.logger.info(Msg.nd_deleted_or_private.format(bad))
        result = self.__bucket
        del self.__bucket
        return result

    async def _get_infos_worker(self, video_id: str):
        async with asyncio.Semaphore(self.__parallel_limit):
            async with self.session.get(URL.URL_Info + video_id) as resp:
                result = await resp.text()

            soup = BeautifulSoup(result, "html.parser")
            if soup.nicovideo_thumb_response["status"].lower() == "ok":
                self.__bucket[video_id] = {
                    KeyGTI.FILE_NAME    : utils.t2filename(soup.select(KeyGTI.TITLE)[0].text),
                    KeyGTI.THUMBNAIL_URL: soup.select(KeyGTI.THUMBNAIL_URL)[0].text,
                    KeyGTI.TITLE        : html.unescape(soup.select(KeyGTI.TITLE)[0].text),
                    KeyGTI.VIDEO_ID     : video_id
                }


class VideoSmile(utils.CanopyAsync):
    def __init__(self,
                 mail: str=None, password: str=None,
                 logger: utils.NTLogger=None,
                 session: aiohttp.ClientSession=None,
                 return_session=False,
                 division: int=4,
                 limit: int=4,
                 chunk_size=1024*50,
                 multiline=True,
                 loop: asyncio.AbstractEventLoop=None,
                 ):
        """
        Smileサーバーから動画をダウンロードする。

        :param mail: メールアドレス
        :param password: パスワード
        :param logger: ロガー
        :param session: セッション
        :param return_session: セッションを返すか
        :param division: いくつに分割するか
        :param limit: (実際のダウンロード前のファイルサイズの確認で)同時にアクセスする最大数
        :param chunk_size: サーバーに一度に要求するデータ量
        :param multiline: プログレスバーを複数行で表示するか
        :param loop: イベントループ
        """
        super().__init__(loop=loop, logger=logger)
        self.__mail = mail
        self.__password = password
        self.__downloaded_size = None  # type: List[int]
        self.__multiline = multiline
        self.__division = division
        self.session = session or self.loop.run_until_complete(self.get_session())
        self.__return_session = return_session
        self.__parallel_limit = limit
        self.__chunk_size = chunk_size

    async def get_session(self) -> aiohttp.ClientSession:
        if self.session:
            return self.session
        else:
            cook = utils.LogIn(mail=self.__mail, password=self.__password).cookie
            return aiohttp.ClientSession(cookies=cook)

    def close(self):
        self.session.close()

    def start(self,
              glossary: Union[list, dict],
              save_dir: Union[str, Path]):
        # TODO Downloading in Economy mode
        self.save_dir = utils.make_dir(save_dir)
        self.logger.debug("Directory to save in: {}".format(self.save_dir))
        # 分割数と同じだけの要素を持つリストを作り、各要素にそれぞれが
        # 保存したファイルサイズを記録する。プログレスバーに利用する。
        self.__downloaded_size = [0] * self.__division

        if isinstance(glossary, list):
            glossary = utils.validator(glossary)
            info = Info(
                mail=self.__mail, password=self.__password,
                session=self.session, return_session=True)
            glossary = info.get_data(glossary)
            self.session = info.session
        self.glossary = glossary
        self.logger.debug("Dictionary of Videos: {}".format(self.glossary))
        self.logger.info(Msg.nd_start_dl_video.format(
            count=len(self.glossary), ids=list(self.glossary)))

        # まず各動画のファイルサイズを集める。
        self.loop.run_until_complete(self._push_file_size())
        self.loop.run_until_complete(self._broker())
        if not self.__return_session:
            self.session.close()
        return self

    async def _push_file_size(self):
        video_ids = sorted(self.glossary)
        tasks = [self._get_file_size_worker(video_id) for video_id in video_ids]
        async with asyncio.Semaphore(self.__parallel_limit):
            result = await asyncio.gather(*tasks)
        for _id, size in zip(video_ids, result):
            self.glossary[_id][KeyDmc.FILE_SIZE] = size

    async def _get_file_size_worker(self, video_id: str) -> int:
        vid_url = self.glossary[video_id][KeyDmc.VIDEO_URL_SM]
        self.logger.debug("Video ID: {}, Video URL: {}".format(video_id, vid_url))
        async with self.session.head(vid_url) as resp:
            headers = resp.headers
            self.logger.debug(str(headers))
            return int(headers["content-length"])

    async def _broker(self):
        futures = []
        for idx, video_id in enumerate(self.glossary):
            coro = self._download(idx, video_id)
            f = asyncio.ensure_future(coro)
            f.add_done_callback(functools.partial(self._combiner, video_id))
            futures.append(f)
        await asyncio.wait(futures, loop=self.loop)

    async def _download(self, idx: int, video_id: str):
        division = self.__division
        file_path = self.make_name(video_id, self.glossary[video_id][KeyDmc.MOVIE_TYPE])

        self.logger.info(
            Msg.nd_download_video.format(
                idx + 1, len(self.glossary), video_id,
                self.glossary[video_id][KeyDmc.TITLE]))

        video_url = self.glossary[video_id][KeyDmc.VIDEO_URL_SM]
        file_size = self.glossary[video_id][KeyDmc.FILE_SIZE]
        headers = [
            {"Range": "bytes={}-{}".format(
                int(file_size*order/division),
                int((file_size*(order+1))/division-1)
            )} for order in range(division)]
        [self.logger.debug(str(h)) for h in headers]

        if self.__multiline:
            progress_bars = [tqdm(total=int(file_size / division),
                                  leave=False, position=order,
                                  unit="B", unit_scale=True,
                                  file=sys.stdout)
                             for order in range(division)]  # type: List[tqdm]
            tasks = [self._download_worker(file_path, video_url, header, order, pbar)
                     for header, order, pbar
                     in zip(headers, range(division), progress_bars)]
            progress_bars = await asyncio.gather(*tasks)  # type: List[tqdm]
            # ネストの「内側」から順に消さないと棒が画面に残る。
            for pbar in reversed(progress_bars):
                pbar.close()
        else:
            tasks = [self._download_worker(file_path, video_url, header, order)
                     for header, order
                     in zip(headers, range(division))]
            await asyncio.gather(*tasks, self._counter_whole(file_size))

    async def _download_worker(self, file_path: Union[str, Path], video_url: str,
                               header: dict, order: int, pbar: tqdm=None) -> tqdm:
        file_path = Path("{}.{:03}".format(file_path, order))
        # => video.mp4.000 ～ video.mp4.003 (4分割の場合)
        with file_path.open("wb") as fd:
            async with self.session.get(url=video_url, headers=header) as video_data:
                self.logger.debug("Started! Header: {}, Video URL: {}".format(header, video_url))
                while True:
                    data = await video_data.content.read(self.__chunk_size)
                    if not data:
                        break
                    downloaded_size = fd.write(data)
                    self.__downloaded_size[order] += downloaded_size
                    if pbar:
                        pbar.update(downloaded_size)
        self.logger.debug("Order: {} Done!".format(order))
        return pbar

    async def _counter_whole(self, file_size: int, interval: int=1):
        """
        ダウンロード済みのファイルサイズを総合して一つのプログレスバーに表示する。

        :param int file_size: 全体のファイルサイズ
        :param int interval: ダウンロード率を更新する間隔
        """
        with tqdm(total=file_size, unit="B") as bar:
            oldsize = 0
            while True:
                newsize = sum(self.__downloaded_size)
                if newsize >= file_size:
                    bar.update(file_size - oldsize)
                    break
                bar.update(newsize - oldsize)
                oldsize = newsize
                await asyncio.sleep(interval)

    def _combiner(self, video_id: str, coroutine: asyncio.Task):
        """
        ダウンロードが終わった後に分割したそれぞれを一つにまとめる関数。

        :param str video_id:
        :param asyncio.Task coroutine: 動画をダウンロードしたタスク
        """
        if coroutine.done() and not coroutine.cancelled():
            file_path = self.make_name(video_id, self.glossary[video_id][KeyDmc.MOVIE_TYPE])
            file_names = ["{}.{:03}".format(file_path, order) for order in range(self.__division)]
            self.logger.debug("File names: {}".format(file_names))
            with file_path.open("wb") as fd:
                for name in file_names:
                    with open(name, "rb") as file:
                        fd.write(file.read())
                    os.remove(name)
            self.logger.info(Msg.nd_download_done.format(path=file_path))


class VideoDmc(utils.CanopyAsync):
    def __init__(self,
                 mail: str=None, password: str=None,
                 logger: utils.NTLogger=None,
                 session: aiohttp.ClientSession=None,
                 return_session=False,
                 division: int=4,
                 chunk_size=1024*50,
                 multiline=True,
                 loop: asyncio.AbstractEventLoop=None,
                 ):
        """
        DMCサーバーから動画をダウンロードする。

        :param mail: メールアドレス
        :param password: パスワード
        :param logger: ロガー
        :param session: セッション
        :param return_session: セッションを返すか
        :param division: いくつに分割するか
        :param chunk_size: 一度にサーバーに要求するデータ量
        :param multiline: プログレスバーを複数行で表示するか
        :param loop: イベントループ
        """
        super().__init__(loop=loop, logger=logger)
        self.__mail = mail
        self.__password = password
        self.__downloaded_size = None  # type: List[int]
        self.__multiline = multiline
        self.__division = division
        self.session = session or self.loop.run_until_complete(self.get_session())
        self.__return_session = return_session
        self.__chunk_size = chunk_size

    async def get_session(self) -> aiohttp.ClientSession:
        if self.session:
            return self.session
        else:
            cook = utils.LogIn(mail=self.__mail, password=self.__password).cookie
            return aiohttp.ClientSession(cookies=cook)

    def close(self):
        self.session.close()

    def start(self,
              glossary: Union[list, dict],
              save_dir: Union[str, Path],
              xml: bool=True):
        self.save_dir = utils.make_dir(save_dir)
        self.__downloaded_size = [0] * self.__division
        self.logger.debug("Directory to save in: {}".format(self.save_dir))

        if isinstance(glossary, list):
            glossary = utils.validator(glossary)
            info = Info(
                mail=self.__mail, password=self.__password,
                session=self.session, return_session=True)
            glossary = info.get_data(glossary)
            self.session = info.session
        self.glossary = glossary
        self.logger.debug("Dictionary of Videos: {}".format(self.glossary))
        self.logger.info(Msg.nd_start_dl_video.format(
            count=len(self.glossary), ids=list(self.glossary)))

        self.loop.run_until_complete(self._broker(xml))
        if not self.__return_session:
            self.session.close()
        return self

    async def _broker(self, xml: bool=True) -> None:
        for idx, video_id in enumerate(self.glossary):
            if self.glossary[video_id][KeyDmc.API_URL] is None:
                self.logger.warning("{} はDMC動画ではありません。従来サーバーの動画を"
                                    "ダウンロードする場合は --smile をコマンドに"
                                    "指定してください。".format(video_id))
                continue
            if xml:
                res_xml = await self._first_nego_xml(video_id)
                video_url = self._extract_video_url_xml(res_xml)
                coro_heartbeat = asyncio.ensure_future(self._heartbeat(video_id, res_xml))
            else:
                res_json = await self._first_nego_json(video_id)
                video_url = self._extract_video_url_json(res_json)
                coro_heartbeat = asyncio.ensure_future(self._heartbeat(video_id, res_json))

            self.logger.debug("動画URL: {}".format(video_url))
            coro_download = asyncio.ensure_future(self._download(idx, video_id, video_url))
            coro_download.add_done_callback(functools.partial(self._canceler, coro_heartbeat))
            coro_download.add_done_callback(functools.partial(self._combiner, video_id))
            tasks = [coro_download, coro_heartbeat]
            await asyncio.gather(*tasks)

    async def _first_nego_xml(self, video_id: str) -> str:
        payload = self._make_param_xml(self.glossary[video_id])
        self.logger.debug("Attempting to first negotiation of {}".format(video_id))
        self.logger.debug("This is the posting XML: {}".format(payload))
        async with self.session.post(
                url=self.glossary[video_id][KeyDmc.API_URL],
                params={"_format": "xml"},
                data=payload,
        ) as response:  # type: aiohttp.ClientResponse
            return await response.text()

    async def _first_nego_json(self, video_id: str) -> str:  # pragma: no cover
        payload = self._make_param_json(self.glossary[video_id])
        self.logger.debug("Attempting to first negotiation of {}".format(video_id))
        self.logger.debug("This is the posting JSON: {}".format(payload))
        async with self.session.post(
                url=self.glossary[video_id][KeyDmc.API_URL],
                params={"_format": "json"},
                data=payload,
        ) as response:  # type: aiohttp.ClientResponse
            return await response.text()

    def _make_param_xml(self, info: Dict[str, str]) -> str:
        info.update({
            "video_src_ids_xml": "".join(map(
                lambda _: "<string>{}</string>".format(_), info[KeyDmc.VIDEO_SRC_IDS])),
            "audio_src_ids_xml": "".join(map(
                lambda _: "<string>{}</string>".format(_), info[KeyDmc.AUDIO_SRC_IDS]))
        })
        xml = Template("""<session>
          <recipe_id>${recipe_id}</recipe_id>
          <content_id>${content_id}</content_id>
          <content_type>movie</content_type>
          <protocol>
            <name>http</name>
            <parameters>
              <http_parameters>
                <method>GET</method>
                <parameters>
                  <http_output_download_parameters>
                    <file_extension>${movie_type}</file_extension>
                  </http_output_download_parameters>
                </parameters>
              </http_parameters>
            </parameters>
          </protocol>
          <priority>${priority}</priority>
          <content_src_id_sets>
            <content_src_id_set>
              <content_src_ids>
                <src_id_to_mux>
                  <video_src_ids>
                    ${video_src_ids_xml}
                  </video_src_ids>
                  <audio_src_ids>
                    ${audio_src_ids_xml}
                  </audio_src_ids>
                </src_id_to_mux>
              </content_src_ids>
            </content_src_id_set>
          </content_src_id_sets>
          <keep_method>
            <heartbeat>
              <lifetime>${heartbeat}</lifetime>
            </heartbeat>
          </keep_method>
          <timing_constraint>unlimited</timing_constraint>
          <session_operation_auth>
            <session_operation_auth_by_signature>
              <token>${token}</token>
              <signature>${signature}</signature>
            </session_operation_auth_by_signature>
          </session_operation_auth>
          <content_auth>
            <auth_type>${auth_type}</auth_type>
            <service_id>nicovideo</service_id>
            <service_user_id>${service_user_id}</service_user_id>
            <max_content_count>10</max_content_count>
            <content_key_timeout>${content_key_timeout}</content_key_timeout>
          </content_auth>
          <client_info>
            <player_id>${player_id}</player_id>
          </client_info>
        </session>
        """)
        return xml.substitute(info)

    def _make_param_json(self, info: Dict[str, Union[str, list, int]]) -> str:  # pragma: no cover
        param = {
            "session": {
                "recipe_id": info[KeyDmc.RECIPE_ID],
                "content_id": info[KeyDmc.CONTENT_ID],
                "content_type": "movie",
                "content_src_id_sets": [
                    {
                        "content_src_ids": [
                            {
                                "src_id_to_mux": {
                                    "video_src_ids": info[KeyDmc.VIDEO_SRC_IDS],
                                    "audio_src_ids": info[KeyDmc.AUDIO_SRC_IDS]
                                }
                            }
                        ]
                    }
                ],
                "timing_constraint": "unlimited",
                "keep_method": {
                    "heartbeat": {
                        "lifetime": info[KeyDmc.HEARTBEAT]
                    }
                },
                "protocol": {
                    "name": "http",
                    "parameters": {
                        "http_parameters": {
                            "parameters": {
                                "http_output_download_parameters": []
                            }
                        }
                    }
                },
                "content_uri": "",
                "session_operation_auth": {
                    "session_operation_auth_by_signature": {
                        "token": info[KeyDmc.TOKEN],
                        "signature": info[KeyDmc.SIGNATURE]
                    }
                },
                "content_auth": {
                    "auth_type": info[KeyDmc.AUTH_TYPE],
                    "max_content_count": 10,
                    "content_key_timeout": info[KeyDmc.C_K_TIMEOUT],
                    "service_id": "nicovideo",
                    "service_user_id": info[KeyDmc.SVC_USER_ID]
                },
                "client_info": {
                    "player_id": info[KeyDmc.PLAYER_ID]
                },
                "priority": info[KeyDmc.PRIORITY]
            }
        }
        result = json.dumps(param)
        return result

    def _extract_video_url_xml(self, text: str) -> str:
        self.logger.debug("Returned XML data: {}".format(text))
        soup = BeautifulSoup(text, "html.parser")
        url_tag = soup.content_uri  # type: Tag
        return url_tag.text

    def _extract_video_url_json(self, text: str) -> str:  # pragma: no cover
        self.logger.debug("Returned JSON data: {}".format(text))
        soup = json.loads(text)
        url_tag = soup["data"]["session"]["content_uri"]
        return url_tag

    def _extract_session_id_xml(self, text: str) -> str:
        soup = BeautifulSoup(text, "html.parser")
        id_tag = soup.session.id  # type: Tag
        self.logger.debug("Session ID: {}".format(id_tag.text))
        return id_tag.text

    def _extract_session_id_json(self, text: str) -> str:  # pragma: no cover
        soup = json.loads(text)
        id_tag = soup["data"]["session"]["id"]
        self.logger.debug("Session ID: {}".format(id_tag))
        return id_tag.text

    def _extract_session_tag(self, text: str) -> str:
        return re.sub(".+(<session>.+</session>).+", r"\1", text)
        # return xml_text[xml_text.find("<session>"): xml_text.find("</session>")+10]

    async def _heartbeat(self, video_id: str, text: str) -> None:
        """
        動画を視聴中に定期的に特定のデータを送信し続ける必要がある。そのための関数。

        :param video_id:
        :param text:
        """
        try:
            self.logger.debug("返ってきたXML: {}".format(text))
            api_url = self.glossary[video_id][KeyDmc.API_URL]
            # 1分ちょうどで送ると遅れるので、待ち時間を少し早める
            waiting = (self.glossary[video_id][KeyDmc.HEARTBEAT] / 1000) - 5
            companion = self._extract_session_tag(text)
            self.logger.debug("送信するXML: {}".format(companion))
            session_id = self._extract_session_id_xml(text)
            await asyncio.sleep(waiting)
            async with self.session.post(
                    url=api_url + "/" + session_id,
                    params={"_format": "xml", "_method": "PUT"},
                    data=companion
            ) as response:  # type: aiohttp.ClientResponse
                res_text = await response.text()
            await self._heartbeat(video_id, res_text)
        except asyncio.CancelledError:
            pass

    async def _get_file_size(self, video_id: str, video_url: str) -> int:
        self.logger.debug("Video ID: {}, Video URL: {}".format(video_id, video_url))
        async with self.session.head(video_url) as resp:
            headers = resp.headers
            self.logger.debug(str(headers))
            return int(headers["content-length"])

    async def _download(self, idx: int, video_id: str, video_url: str):
        division = self.__division
        file_path = self.make_name(video_id, self.glossary[video_id][KeyDmc.MOVIE_TYPE])

        self.logger.info(
            Msg.nd_download_video.format(
                idx + 1, len(self.glossary), video_id,
                self.glossary[video_id][KeyDmc.TITLE]))

        file_size = await self._get_file_size(video_id, video_url)
        headers = [
            {"Range": "bytes={}-{}".format(
                int(file_size*order/division),
                int((file_size*(order+1))/division-1)
            )} for order in range(division)]
        [self.logger.debug("Order: {}, {}".format(o, h)) for o, h in zip(range(division), headers)]

        if self.__multiline:
            progress_bars = [tqdm(total=int(file_size / division),
                                  leave=False, position=order,
                                  unit="B", unit_scale=True,
                                  file=sys.stdout)
                             for order in range(division)]  # type: List[tqdm]
            tasks = [self._download_worker(file_path, video_url, header, order, pbar)
                     for header, order, pbar
                     in zip(headers, range(division), progress_bars)]
            progress_bars = await asyncio.gather(*tasks)  # type: List[tqdm]
            # ネストの「内側」から順に消さないと棒が画面に残る。
            for pbar in reversed(progress_bars):
                pbar.close()
        else:
            tasks = [self._download_worker(file_path, video_url, header, order)
                     for header, order
                     in zip(headers, range(division))]
            await asyncio.gather(*tasks, self._counter_whole(file_size))

    async def _download_worker(self, file_path: Union[str, Path], video_url: str,
                               header: dict, order: int, pbar: tqdm = None) -> tqdm:
        file_path = Path("{}.{:03}".format(file_path, order))
        # => video.mp4.000 ～ video.mp4.003 (4分割の場合)
        self.logger.debug(file_path)
        with file_path.open("wb") as fd:
            async with self.session.get(url=video_url, headers=header) as video_data:
                self.logger.debug("Started! Header: {}, Video URL: {}".format(header, video_url))
                while True:
                    data = await video_data.content.read(self.__chunk_size)
                    if not data:
                        break
                    downloaded_size = fd.write(data)
                    self.__downloaded_size[order] += downloaded_size
                    if pbar:
                        pbar.update(downloaded_size)
        self.logger.debug("Order: {} Done!".format(order))
        return pbar

    def _canceler(self, task_to_cancel: asyncio.Task, _: asyncio.Task) -> bool:
        """
        動画のダウンロードが終わった後にHeartbeatを止めるための関数。

        :param asyncio.Task task_to_cancel:
        :param asyncio.Task _:
        :rtype: bool
        """
        return task_to_cancel.cancel()

    async def _counter_whole(self, file_size: int, interval: int=1):
        with tqdm(total=file_size, unit="B") as bar:
            oldsize = 0
            while True:
                newsize = sum(self.__downloaded_size)
                if newsize >= file_size:
                    bar.update(file_size - oldsize)
                    break
                bar.update(newsize - oldsize)
                oldsize = newsize
                await asyncio.sleep(interval)

    def _combiner(self, video_id: str, coroutine: asyncio.Task):
        """
        ダウンロードが終わった後に分割したそれぞれを一つにまとめる関数。

        :param str video_id:
        :param asyncio.Task coroutine:
        """
        if coroutine.done() and not coroutine.cancelled():
            file_path = self.make_name(video_id, self.glossary[video_id][KeyDmc.MOVIE_TYPE])
            file_names = ["{}.{:03}".format(file_path, order) for order in range(self.__division)]
            self.logger.debug("File names: {}".format(file_names))
            with file_path.open("wb") as fd:
                for name in file_names:
                    with open(name, "rb") as file:
                        fd.write(file.read())
                    os.remove(name)
            self.logger.info(Msg.nd_download_done.format(path=file_path))


class Comment(utils.CanopyAsync):
    def __init__(self,
                 mail: str=None, password: str=None,
                 logger: utils.NTLogger=None,
                 session: aiohttp.ClientSession=None,
                 return_session=False,
                 limit: int=4,
                 wayback=False,
                 loop: asyncio.AbstractEventLoop=None,
                 ):
        """
        コメントをダウンロードする。

        :param mail: メールアドレス
        :param password: パスワード
        :param logger: ロガー
        :param session: セッション
        :param return_session: セッションを返すか
        :param limit: 同時にアクセスする最大数
        :param wayback: 過去ログを取りに行くかどうか
        :param loop: イベントループ
        """
        super().__init__(loop=loop, logger=logger)
        self.__mail = mail
        self.__password = password
        self.__downloaded_size = None  # type: List[int]
        self.session = session or self.loop.run_until_complete(self.get_session())
        self.__return_session = return_session
        self.__parallel_limit = limit
        self.__wayback = wayback

    async def get_session(self) -> aiohttp.ClientSession:
        if self.session:
            return self.session
        else:
            cook = utils.LogIn(mail=self.__mail, password=self.__password).cookie
            return aiohttp.ClientSession(cookies=cook)

    def close(self):
        self.session.close()

    def start(self, glossary, save_dir, xml=False, density: str="0-99999:9999,1000"):
        """
        ダウンロードを開始する。

        0-99999:9999,1000: 「0分～99999分までの範囲で
        一分間あたり9999件、直近の1000件を取得する」の意味。

        :param dict[str, dict[str, int | str]] | list[str] glossary:
        :param str | Path save_dir:
        :param bool xml:
        :param str density: ダウンロードするコメントの密度。
        """
        utils.check_arg(locals())
        self.save_dir = utils.make_dir(save_dir)
        self.logger.debug("Directory to save in: {}".format(self.save_dir))

        if isinstance(glossary, list):
            glossary = utils.validator(glossary)
            info = Info(
                mail=self.__mail, password=self.__password,
                session=self.session, return_session=True)
            glossary = info.get_data(glossary)
            self.session = info.session
        self.glossary = glossary

        self.logger.info(Msg.nd_start_dl_comment.format(
            count=len(self.glossary), ids=list(self.glossary)))

        futures = []
        for idx, video_id in enumerate(self.glossary):
            coro = self._download(idx, self.glossary[video_id], xml, density)
            f = asyncio.ensure_future(coro)
            f.add_done_callback(functools.partial(self.saver, video_id, xml))
            futures.append(f)

        self.loop.run_until_complete(asyncio.wait(futures, loop=self.loop))
        return self

    async def _download(self, idx: int, info: dict, is_xml: bool, density: str) -> str:
        utils.check_arg(locals())

        video_id        = info[KeyDmc.VIDEO_ID]
        thread_id       = info[KeyDmc.THREAD_ID]
        msg_server      = info[KeyDmc.MSG_SERVER]
        user_id         = info[KeyDmc.USER_ID]
        user_key        = info[KeyDmc.USER_KEY]

        # 以下は公式動画で必要
        opt_thread_id   = info[KeyDmc.OPT_THREAD_ID]    # int なければ None
        needs_key       = info[KeyDmc.NEEDS_KEY]        # int なければ None
        thread_key      = None
        force_184       = None

        self.logger.info(
            Msg.nd_download_comment.format(
                idx + 1, len(self.glossary), video_id,
                info[KeyGTI.TITLE]))


        is_official = re.match("^(?:so|\d)", info[KeyDmc.VIDEO_ID]) is not None

        if is_official:
            thread_key, force_184 = await self.get_thread_key(thread_id, needs_key)

        if self.__wayback:
            waybackkey = await self.get_wayback_key(thread_id)

        if is_xml:
            req_param = self.make_param_xml(
                thread_id, user_id, thread_key, force_184, density=density)
            com_data = await self.retriever(data=req_param, url=msg_server)
        else:
            req_param = self.make_param_json(
                is_official, user_id, user_key, thread_id,
                opt_thread_id, thread_key, force_184, density=density)
            com_data = await self.retriever(data=json.dumps(req_param), url=URL.URL_Msg_JSON)

        return self.postprocesser(is_xml, com_data)

    async def retriever(self, data: str, url: str) -> str:
        self.logger.debug("Posting Parameters: {}".format(data))
        async with asyncio.Semaphore(self.__parallel_limit):
            async with self.session.post(url=url, data=data) as resp:  # type: aiohttp.ClientResponse
                return await resp.text()

    def postprocesser(self, is_xml: bool, result: str):
        """
        取ってきたコメントデータに後処理する。

        :param bool is_xml: 受け取ったデータがXML形式かどうか。
        :param str result: コメントの文字列
        :rtype: str
        """
        if is_xml:
            return result.replace("><", ">\n<")
        else:
            return result.replace("}, ", "},\n")

    def saver(self, video_id: str, is_xml: bool, coroutine: asyncio.Task) -> bool:
        utils.check_arg(locals())
        comment_data = coroutine.result()
        if is_xml:
            extention = "xml"
        else:
            extention = "json"

        file_path = self.make_name(video_id, extention)
        self.logger.debug("File Path: {}".format(file_path))
        with file_path.open("w", encoding="utf-8") as f:
            f.write(comment_data + "\n")
        self.logger.info(Msg.nd_download_done.format(path=file_path))
        return True

    async def get_thread_key(self, thread_id, needs_key):
        """
        専用のAPIにアクセスして thread_key を取得する。

        :param str thread_id:
        :param str needs_key:
        :rtype: tuple[str, str]
        """
        utils.check_arg(locals())
        if not int(needs_key) == 1:
            self.logger.debug("needs_key is not 1. Video ID (or Thread ID): {},"
                              " needs_key: {}".format(thread_id, needs_key))
            return "", "0"
        async with self.session.get(URL.URL_GetThreadKey, params={"thread": thread_id}) as resp:
            response = await resp.text()
        self.logger.debug("Response from GetThreadKey API"
                          " (thread id is {}): {}".format(thread_id, response))
        parameters = parse_qs(response)
        threadkey = parameters["threadkey"][0]  # type: str
        force_184 = parameters["force_184"][0]  # type: str
        return threadkey, force_184

    async def get_wayback_key(self, thread_id: int):
        async with asyncio.Semaphore(self.__parallel_limit):
            async with self.session.get(URL.URL_WayBackKey, params={"thread", thread_id}) as resp:
                response = await resp.text()
                self.logger.debug("Waybackkey response: {}".format(response))
            return parse_qs(response)["waybackkey"][0]

    def make_param_xml(self, thread_id, user_id, thread_key=None, force_184=None,
                       waybackkey=None, quantity=1000, density="0-99999:9999,1000"):
        """
        コメント取得用のxmlを構成する。

        fork="1" があると投稿者コメントを取得する。
        0-99999:9999,1000: 「0分～99999分までの範囲で
        一分間あたり9999件、直近の1000件を取得する」の意味。

        :param str thread_id:
        :param str user_id:
        :param str thread_key:
        :param str force_184:
        :param str waybackkey:
        :param int | str quantity:取りに行くコメント数
        :param str density: 取りに行くコメントの密度。 0-99999:9999,1000 のような形式。
        :rtype: str
        """
        utils.check_arg({"thread_id": thread_id, "user_id": user_id})
        self.logger.debug("Arguments: {}".format(locals()))
        wbk = 'waybackkey="{}"'.format(waybackkey) if waybackkey else ""
        if thread_key:
            return (
                '<packet>'
                '<thread thread="{t_id}" user_id="{user_id}" scores="1"'
                ' threadkey="{t_key}" force_184="{force}"'
                ' {wbk} version="20090904" res_from="-{quantity}"/>'
                '<thread thread="{t_id}" user_id="{user_id}" scores="1"'
                ' threadkey="{t_key}" force_184="{force}"'
                ' {wbk} version="20090904" res_from="-{quantity}" fork="1"/>'
                '<thread_leaves thread="{t_id}" user_id="{user_id}" scores="1">'
                '{density}</thread_leaves>'
                '</packet>').format(t_id=thread_id, user_id=user_id,
                                    t_key=thread_key, force=force_184, wbk=wbk,
                                    quantity=quantity, density=density)
        else:
            return (
                '<packet>'
                '<thread thread="{t_id}" user_id="{user_id}" scores="1"'
                ' {wbk} version="20090904" res_from="-{quantity}"/>'
                '<thread thread="{t_id}" user_id="{user_id}" scores="1"'
                ' {wbk} version="20090904" res_from="-{quantity}" fork="1"/>'
                '<thread_leaves thread="{t_id}" user_id="{user_id}" scores="1">'
                '{density}</thread_leaves>'
                '</packet>').format(t_id=thread_id, user_id=user_id, wbk=wbk,
                                    quantity=quantity, density=density)

    def make_param_json(self, official_video, user_id, user_key, thread_id,
                        optional_thread_id=None, thread_key=None, force_184=None,
                        density="0-99999:9999,1000"):
        """
        コメント取得用のjsonを構成する。

        fork="1" があると投稿者コメントを取得する。
        0-99999:9999,1000: 「0分～99999分までの範囲で
        一分間あたり9999件、直近の1000件を取得する」の意味。

        :param bool official_video: 公式動画なら True
        :param str user_id:
        :param str user_key:
        :param str thread_id:
        :param str | None optional_thread_id:
        :param str | None thread_key:
        :param str | None force_184:
        :param str density: 取りに行くコメントの密度。 0-99999:9999,1000 のような形式。
        :rtype: str
        """
        utils.check_arg({"official_video": official_video, "user_id": user_id,
                         "user_key": user_key, "thread_id": thread_id})
        self.logger.debug("Arguments of creating JSON: {}".format(locals()))
        result = [
            {"ping": {"content": "rs:0"}},
            {"ping": {"content": "ps:0"}},
            {
                "thread": {
                    "thread"     : optional_thread_id or thread_id,
                    "version"    : "20090904",
                    "language"   : 0,
                    "user_id"    : user_id,
                    "with_global": 1,
                    "scores"     : 1,
                    "nicoru"     : 0,
                    "userkey"    : user_key
                }
            },
            {"ping": {"content": "pf:0"}},
            {"ping": {"content": "ps:1"}},
            {
                "thread_leaves": {
                    "thread"  : optional_thread_id or thread_id,
                    "language": 0,
                    "user_id" : user_id,
                    # "content" : "0-4:100,250",  # 公式仕様のデフォルト値
                    "content" : density,
                    "scores"  : 1,
                    "nicoru"  : 0,
                    "userkey" : user_key
                }
            },
            {"ping": {"content": "pf:1"}}
        ]

        if official_video:
            result += [{"ping": {"content": "ps:2"}},
                       {
                           "thread": {
                               "thread"     : thread_id,
                               "version"    : "20090904",
                               "language"   : 0,
                               "user_id"    : user_id,
                               "force_184"  : force_184,
                               "with_global": 1,
                               "scores"     : 1,
                               "nicoru"     : 0,
                               "threadkey"  : thread_key
                           }
                       },
                       {"ping": {"content": "pf:2"}},
                       {"ping": {"content": "ps:3"}},
                       {
                           "thread_leaves": {
                               "thread"   : thread_id,
                               "language" : 0,
                               "user_id"  : user_id,
                               # "content"  : "0-4:100,250",  # 公式仕様のデフォルト値
                               "content"  : density,
                               "scores"   : 1,
                               "nicoru"   : 0,
                               "force_184": force_184,
                               "threadkey": thread_key
                           }
                       },
                       {"ping": {"content": "pf:3"}}]
        result += [{"ping": {"content": "rf:0"}}]
        return result


def main(args):
    """
    メイン。

    :param args: ArgumentParser.parse_args() によって解釈された引数
    :rtype: bool
    """
    is_debug = int(os.getenv("PYTHON_TEST", 0))
    mailadrs = args.mail[0] if args.mail else None
    password = args.password[0] if args.password else None

    """ エラーの除外 """
    videoid = utils.validator(args.VIDEO_ID)
    if not videoid:
        sys.exit(Err.invalid_videoid)
    if not (args.getthumbinfo or args.thumbnail or args.comment or args.video):
        sys.exit(Err.not_specified.format("--thumbnail、 --comment、 --video のいずれか"))

    if args.getthumbinfo:
        file_name = args.out[0] if isinstance(args.out, list) else None
        utils.print_info(videoid, file_name)
        sys.exit()

    """ 本筋 """
    log_level = "DEBUG" if is_debug else args.loglevel
    logger = utils.NTLogger(log_level=log_level, file_name=utils.LOG_FILE_ND)
    destination = utils.make_dir(args.dest[0])

    if args.thumbnail:
        Thumbnail(logger=logger).start(videoid, destination)
        if not (args.comment or args.video):
            # サムネイルのダウンロードだけなら
            # ログインする必要がないのでここで終える。
            return True

    info = Info(
        mail=mailadrs, password=password,
        logger=logger, sieve=args.nosieve, return_session=True)
    database = info.get_data(videoid)
    session = info.session

    if len(database) == 0:
        session.close()
        return True

    if args.comment:
        (Comment(logger=logger, session=session)
         .start(database, destination, args.xml))

    if args.video:
        if args.smile:
            (VideoSmile(logger=logger, session=session,
                        division=args.limit, multiline=args.nomulti)
             .start(database, destination))
        else:
            (VideoDmc(logger=logger, session=session,
                      division=args.limit, multiline=args.nomulti)
             .start(database, destination))

    if not session.closed:
        session.close()
    return True



# coding: UTF-8
import os
import sys

from . import nicodown, nicoml
from .utils import Msg, Err, InheritedParser

if sys.version_info >= (3, 5):
    from . import nicodown_async, nicoml_async
else:
    nicodown_async, nicoml_async = None, None


def main(arguments=None, async_=True):
    """
    メイン。

    :param arguments: 引数の文字列
    :param bool async_:
    :rtype: bool
    """
    choices = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]

    parser = InheritedParser(prog="nicotools", fromfile_prefix_chars="+", description=Msg.description)
    # nargs があると値はリストに入る。
    parser.add_argument("-l", "--mail", nargs=1, help=Msg.nd_help_mail, metavar="MAIL")
    parser.add_argument("-p", "--pass", nargs=1, help=Msg.nd_help_password, metavar="PASSWORD", dest="password")
    parser.add_argument("-w", "--what", action="store_true", help=Msg.nd_help_what)
    parser.add_argument("--loglevel", type=str.upper, default="INFO", help=Msg.nd_help_loglevel, choices=choices)
    subparsers = parser.add_subparsers()


    parser_nd = subparsers.add_parser("download", aliases=["d"], help=Msg.nd_description)
    if async_ and nicodown_async is not None:
        parser_nd.set_defaults(func=nicodown_async.main)
    else:
        parser_nd.set_defaults(func=nicodown.main)
    parser_nd.add_argument("VIDEO_ID", nargs="+", type=str, help=Msg.nd_help_video_id)
    parser_nd.add_argument("--loglevel", type=str.upper, default="INFO", help=Msg.nd_help_loglevel, choices=choices)
    parser_nd.add_argument("-w", "--what", action="store_true", help=Msg.nd_help_what)
    parser_nd.add_argument("-l", "--mail", nargs=1, help=Msg.nd_help_mail, metavar="MAIL")
    parser_nd.add_argument("-p", "--pass", nargs=1, help=Msg.nd_help_password, metavar="WORD", dest="password")
    parser_nd.add_argument("-d", "--dest", nargs=1, type=str, default=[os.getcwd()], help=Msg.nd_help_destination)
    parser_nd.add_argument("-c", "--comment", action="store_true", help=Msg.nd_help_comment)
    parser_nd.add_argument("-v", "--video", action="store_true", help=Msg.nd_help_video)
    parser_nd.add_argument("-t", "--thumbnail", action="store_true", help=Msg.nd_help_thumbnail)
    parser_nd.add_argument("-i", "--getthumbinfo", action="store_true", help=Msg.nd_help_info)
    parser_nd.add_argument("-x", "--xml", action="store_true", help=Msg.nd_help_xml)
    parser_nd.add_argument("-o", "--out", nargs=1, help=Msg.nd_help_outfile, metavar="FILE")
    if async_ and nicodown_async is not None:
        parser_nd.add_argument("--smile", action="store_true", help=Msg.nd_help_smile)
        parser_nd.add_argument("--dmc", action="store_true", help=Msg.nd_help_dmc, default=True)
        parser_nd.add_argument("--limit", type=int, help=Msg.nd_help_limit, default=4)
        parser_nd.add_argument("--nomulti", action="store_false", help=Msg.nd_help_nomulti, dest="nomulti")
        parser_nd.add_argument("--nosieve", action="store_false", help=Msg.nd_help_sieve)


    parser_ml = subparsers.add_parser("mylist", aliases=["m"], help=Msg.ml_description)
    if async_ and nicoml_async is not None:
        parser_ml.set_defaults(func=nicoml_async.main)
    else:
        parser_ml.set_defaults(func=nicoml.main)
    parser_ml.add_argument("src", nargs=1, help=Msg.ml_help_src, metavar="マイリスト名")
    parser_ml.add_argument("--loglevel", type=str.upper, default="INFO", help=Msg.nd_help_loglevel, choices=choices)
    parser_ml.add_argument("-w", "--what", action="store_true", help=Msg.nd_help_what)
    parser_ml.add_argument("-l", "--mail", nargs=1, help=Msg.nd_help_mail, metavar="MAIL")
    parser_ml.add_argument("-p", "--pass", nargs=1, help=Msg.nd_help_password, metavar="WORD", dest="password")
    parser_ml.add_argument("-i", "--id", action="store_true", help=Msg.ml_help_id)
    parser_ml.add_argument("-o", "--out", nargs=1, help=Msg.ml_help_outfile, metavar="FILE")
    parser_ml.add_argument("--yes", action="store_true", help=Msg.ml_help_yes)
    if async_ and nicoml_async is not None:
        parser_ml.add_argument("--each", action="store_true", help=Msg.ml_help_each)

    group_one = parser_ml.add_argument_group(Msg.ml_help_group_a)
    group_one.add_argument("-t", "--to", nargs=1, help=Msg.ml_help_to, metavar="To")
    group_one.add_argument("-a", "--add", nargs="+", help=Msg.ml_help_add, metavar="sm...")
    group_one.add_argument("-d", "--delete", nargs="+", help=Msg.ml_help_delete, metavar="sm...")
    group_one.add_argument("-m", "--move", nargs="+", help=Msg.ml_help_move, metavar="sm...")
    group_one.add_argument("-c", "--copy", nargs="+", help=Msg.ml_help_copy, metavar="sm...")

    group_two = parser_ml.add_argument_group(Msg.ml_help_group_b)
    group_two.add_argument("-r", "--create", action="store_true", help=Msg.ml_help_create)
    group_two.add_argument("--purge", action="store_true", help=Msg.ml_help_purge)
    group_two.add_argument("-s", "--show", action="count", help=Msg.ml_help_show)
    group_two.add_argument("-e", "--export", action="count", help=Msg.ml_help_export)
    group_two.add_argument("--everything", action="store_true", help=Msg.ml_help_everything)

    args = parser.parse_args(globals().get("DEBUG_ARGS") or arguments)
    if (len(sys.argv) <= 1 or not hasattr(args, "func")) and not int(os.getenv("PYTHON_TEST", 0)):
        parser.print_help() or sys.exit()
    if args.what:
        print(args) or sys.exit()

    try:
        return args.func(args)
    except KeyboardInterrupt:
        sys.exit(Err.keyboard_interrupt)

if __name__ == "__main__":
    main()



# coding: UTF-8
import asyncio
import inspect
import logging
import os
import re
import sys
from argparse import ArgumentParser
from getpass import getpass
from os.path import join, expanduser
from pathlib import Path
from urllib.parse import parse_qs

import aiohttp
import requests
from requests import cookies

ALL_ITEM = "*"
DEFAULT_NAME = "とりあえずマイリスト"
DEFAULT_ID = 0
LOG_FILE_ND = "nicotools_download.log"
LOG_FILE_ML = "nicotools_mylist.log"
IS_DEBUG = int(os.getenv("PYTHON_TEST", 0))
if IS_DEBUG:
    __os_name = os.getenv("TRAVIS_OS_NAME", os.name)
    __version = (os.getenv("TRAVIS_PYTHON_VERSION") or
               os.getenv("PYTHON_VERSION") or
               "_" .join(map(str, sys.version_info[0:3])))
    COOKIE_FILE_NAME = "nicotools_cokkie_{0}_{1}.txt".format(__os_name, __version)
else:
    COOKIE_FILE_NAME = "nicotools_cookie.txt"

# 文字列をUTF-8以外にエンコードするとき、変換不可能な文字をどう扱うか
BACKSLASH = "backslashreplace"
CHARREF = "xmlcharrefreplace"
REPLACE = "replace"


def get_encoding():
    """
    コンソールの文字コードを返す。sys.stdout.encoding が None の時には UTF-8を返す。

    :rtype: str
    """
    return sys.stdout.encoding or "UTF-8"


def print_info(queue, file_name=None):
    """
    GetThumbInfo にアクセスして返ってきたXMLをそのまま表示する。

    :param list queue:
    :param str | Path | None file_name:
    :return: bool
    """
    text = "\n\n".join([requests.get(URL.URL_Info + video_id).text for video_id in queue])
    if file_name:
        file_name = make_dir(file_name)
        with file_name.open(encoding="utf-8", mode="w") as fd:
            fd.write(text + "\n")
    else:
        print(text.encode(get_encoding(), BACKSLASH).decode(get_encoding()))
    return True


def validator(input_list):
    """
    動画IDが適切なものか確認する。元のリスト内の順番は保持されない。

    受け入れるのは以下の形式:
        * "*"
        * http://www.nicovideo.jp/watch/sm123456
        * http://nico.ms/sm123456
        * sm1234
        * watch/sm123456
        * nm1234
        * watch/nm123456
        * so1234
        * watch/so123456
        * 123456
        * watch/123456

    :param list[str] | tuple[str] | set[str] input_list:
    :rtype: list[str]
    """
    check_arg(locals())
    matcher = re.compile(
        """\s*(?:
        {0}|  # 「全て」を指定するときの記号
        (?:
            (?:(?:h?t?tp://)?www\.nicovideo\.jp/)?watch/  # 通常URL
           |(?:h?t?tp://)?nico\.ms/  # 短縮URL
        )?
            ((?:sm|nm|so)?\d+)  # ID本体
        )\s?""".format(re.escape(ALL_ITEM)), re.I + re.X).match
    if isinstance(input_list, str):
        input_list = [input_list]

    if not isinstance(input_list, (list, tuple, set)):
        raise MylistArgumentError(Err.invalid_argument.format(input_list))

    if "\t" in input_list[0]:
        for line in input_list[1:]:
            if "\t" not in line:
                return []
        input_list = [item.split("\t")[0] for item in input_list]
    else:
        if len(input_list) == 1 and input_list[0] == ALL_ITEM:
            return input_list
        for item in input_list:
            if not matcher(item):
                return []

    # ................................↓"*" が入っていたときの対策
    return [matcher(item).group(1) or item.strip()
            for item in set(input_list) if matcher(item) or ALL_ITEM in item]


def make_dir(directory):
    """
    保存場所に指定されたフォルダーがない場合にはつくり、その絶対パスを返す。

    :param str | Path directory: フォルダー名
    :rtype: Path
    """
    check_arg(locals())
    if isinstance(directory, str):
        directory = Path(directory)
    if directory.suffix:
        return make_dir(directory.parent) / directory.name
    try:
        if not directory.is_dir():
            directory.mkdir(parents=True)
        return directory.resolve()
    except (
            # [WinError 87] パラメーターが間違っています。
            # (con とか nul とか)(Python 3.5以下で起こる。)
            OSError,
            # [WinError 2] 指定されたファイルが見つかりません。(Python 3.5以下で起こる。)
            FileNotFoundError,
            # [WinError 267] ディレクトリ名が無効です。(con とか nul とか)
            NotADirectoryError,
            # [WinError 5] アクセスが拒否されました。(C:/ とか D:/ とか)
            # [Errno 13] Permission denied (同名のフォルダーがあってそこに書き込もうとした場合)
            PermissionError,
            # [WinError 183] 既に存在するファイルを作成することはできません。
            FileExistsError
    ):
        raise NameError(Err.invalid_dirname.format(directory))


def t2filename(text):
    """
    ファイル名に使えない文字を全角文字に置き換える。

    :param str text: ファイル名
    :rtype: str
    """
    mydic = {
        r"\/": "／", "/": "／", "'": "’", "\"": "”",
        "<"  : "＜", ">": "＞", "|": "｜", ":": "：",
        "*"  : "＊", "?": "？", "~": "～", "\\": "＼"
    }
    for item in mydic.keys():
        text = text.replace(item, mydic[item])
    # 置き換えるペアが増えたらこっちを使うと楽かもしれない
    # pattern = re.compile("|".join(re.escape(key) for key in mydic.keys()))
    # return pattern.sub(lambda x: mydic[x.group()], text)
    return text


def check_arg(parameters):
    """
    None を許容しない引数が None を含んでいないか検査する。

    :param dict[str, Object] parameters: パラメーターの名前と値の辞書
    :rtype: None
    """
    for _name, _value in parameters.items():
        if _value is None:
            raise MylistArgumentError(Err.not_specified.format(_name))


def sizeof_fmt(num):
    """
    数字を読みやすい単位で表す。

    :param int num: 計りたい整数
    :rtype: str
    """
    for unit in ['B', 'Kb', 'Mb']:
        if num < 1024.0 and isinstance(num, int):
            return "{:3}{:s}".format(num, unit)
        elif num < 1024.0:
            return "{:3.2f}{:s}".format(num, unit)
        num /= 1024.0
    return "{:.2f}Gb".format(num)


def get_from_getflv(video_id, session, logger=None):
    """
    GetFlv APIから情報を得る。

    * GetFlvのサンプル:

    thread_id=1406370428
    &l=314
    &url=http%3A%2F%2Fsmile-pom32.nicovideo.jp%2Fsmile%3Fm%3D24093152.45465
    &ms=http%3A%2F%2Fmsg.nicovideo.jp%2F27%2Fapi%2F
    &ms_sub=http%3A%2F%2Fsub.msg.nicovideo.jp%2F27%2Fapi%2F
    &user_id=<ユーザーIDの数字>
    &is_premium=1
    &nickname=<URLエンコードされたユーザー名の文字列>
    &time=1475176067845
    &done=true
    &ng_rv=220
    &userkey=1475177867.%7E1%7EhPBJrVv78e251OPzyAiSs1fYAJhYIzDPOq5LNiNqZxs

    * 但しアクセス制限がかかったときには:

    error=access_locked&done=true

    :param str video_id:
    :param requests.Session session:
    :param NTLogger logger:
    :rtype: dict[str, str] | None
    """
    check_arg(locals())
    suffix = {"as3": 1} if video_id.startswith("nm") else None
    response = session.get(URL.URL_GetFlv + video_id, params=suffix)
    if logger: logger.debug("GetFLV Response: {}".format(response.text))
    return extract_getflv(response.text)


def extract_getflv(content):
    """

    :param str content: GetFLV の返事
    :rtype: dict[str, str] | None
    """
    parameters = parse_qs(content)
    if parameters.get("error") is not None:
        return None
    result = {
        KeyGetFlv.THREAD_ID    : int(parameters[KeyGetFlv.THREAD_ID][0]),
        KeyGetFlv.LENGTH       : int(parameters[KeyGetFlv.LENGTH][0]),
        KeyGetFlv.VIDEO_URL    : parameters[KeyGetFlv.VIDEO_URL][0],
        KeyGetFlv.MSG_SERVER   : parameters[KeyGetFlv.MSG_SERVER][0],
        KeyGetFlv.MSG_SUB      : parameters[KeyGetFlv.MSG_SUB][0],
        KeyGetFlv.USER_ID      : int(parameters[KeyGetFlv.USER_ID][0]),
        KeyGetFlv.IS_PREMIUM   : int(parameters[KeyGetFlv.IS_PREMIUM][0]),
        KeyGetFlv.NICKNAME     : parameters[KeyGetFlv.NICKNAME][0],
        KeyGetFlv.USER_KEY     : parameters[KeyGetFlv.USER_KEY][0],

        # 以下は公式動画にだけあるもの。通常の動画ではNone
        KeyGetFlv.OPT_THREAD_ID: None,
        KeyGetFlv.NEEDS_KEY    : None,
    }

    opt_thread_id   = parameters.get(KeyGetFlv.OPT_THREAD_ID, [None])[0]
    needs_key       = parameters.get(KeyGetFlv.NEEDS_KEY, [None])[0]
    if opt_thread_id is not None:
        result.update({
            KeyGetFlv.OPT_THREAD_ID : int(opt_thread_id),
            KeyGetFlv.NEEDS_KEY     : int(needs_key),
        })
    return result


class MylistError(Exception):
    """ マイリスト操作で誤りがあったときに発生させるエラー """
    pass


class MylistNotFoundError(MylistError):
    """ マイリストが見つからなかったときに発生させるエラー """
    pass


class MylistArgumentError(MylistError):
    """ 引数が誤っていたときに発生させるエラー """
    pass


class MylistAPIError(MemoryError):
    """ APIの操作の結果が好ましくない場合に発生させるエラー """
    def __init__(self, code=None, msg=None, ok=False):
        """
        :param str code: APIのエラーコード(の文字列)
        :param str msg: 伝えたい文言
        :param bool ok: 作業を続行してよいかどうか
        """
        self.code = code
        self.msg = msg
        self.ok = ok


class NotLoginError(Exception):
    """ ログインしていない、できなかったときに発生させるエラー """
    pass


class Canopy:
    def __init__(self, logger=None):
        self.glossary = None
        self.save_dir = None  # type: Path
        self.logger = self.get_logger(logger)  # type: NTLogger

    def make_name(self, video_id, ext):
        """
        ファイル名を返す。

        :param str video_id:
        :param str ext:
        :rtype: Path
        """
        check_arg(locals())
        file_name =  Msg.nd_file_name.format(
            vid=video_id, ext=ext,
            name=self.glossary[video_id][KeyGTI.FILE_NAME])
        return Path(self.save_dir).resolve() / file_name

    def get_logger(self, logger):
        """
        ロガーを返す。

        すでにあるならそれを使い、無い、またはハンドラーを持たなければ新しく作って返す。

        :param NTLogger logger:
        :rtype: NTLogger
        """
        if not (logger and hasattr(logger, "handlers")):
            return NTLogger()
        else:
            return logger


class CanopyAsync(Canopy):
    def __init__(self, loop: asyncio.AbstractEventLoop=None, logger=None):
        super().__init__(logger=logger)
        self.loop = loop or asyncio.get_event_loop()
        self.glossary = None
        self.save_dir = None  # type: Path
        self.logger = self.get_logger(logger)  # type: NTLogger
        self.session = None  # type: aiohttp.ClientSession


class LogIn:
    __singleton__ = None
    is_login = False
    cookie = {}

    @classmethod
    def __new__(cls, *more, **kwargs):
        """
        ログイン処理が一度だけなのを保障するためにシングルトンとして振る舞わせる。

        参考:
            デザインパターン（Design Pattern）#Singleton - Qiita
            http://qiita.com/nirperm/items/af1f83925ba43dbf22eb
        """

        if cls.__singleton__ is None:
            cls.__singleton__ = super().__new__(cls)
        return cls.__singleton__

    def __init__(self, mail=None, password=None, session=None):
        """
        :param str | None mail: メールアドレス
        :param str | None password: パスワード
        :param requests.Session | None session: セッションオブジェクト
        """
        if not (session or mail or password):
            self.session = self.get_session()
        elif session and not (mail or password):
            self.session = session
            self.token = self.get_token(self.session)
        else:
            self.session = self.get_session(self.ask_credentials(mail=mail, password=password))

    def get_session(self, auth=None):
        """
        クッキーを読み込み、必要ならばログインし、そのセッションを返す。

        :param dict[str, str] | None auth:
        :rtype: requests.Session
        """

        session = requests.session()
        cook = self.load_cookies()
        if auth or cook:
            if cook:
                session.cookies = requests.cookies.cookiejar_from_dict(cook)
            else:
                session.post(URL.URL_LogIn, params=auth)
            self.token = self.get_token(session)
            if self.token:
                # if self._we_have_logged_in(res.text):
                self.cookie = self.save_cookies(session.cookies)
                self.is_login = True
            else:
                return self.get_session(self.ask_credentials())
        else:
            return self.get_session(self.ask_credentials())
        return session

    def get_token(self, session):
        """
        マイリストの操作に必要な"NicoAPI.token"を取ってくる。

        :param requests.Session session:
        :rtype: str
        """
        htmltext = session.get(URL.URL_MyListTop).text
        try:
            fragment = htmltext.split("NicoAPI.token = \"")[1]
            return fragment[:fragment.find("\"")]
        except IndexError:
            return None

    @classmethod
    def ask_credentials(cls, mail=None, password=None):
        """
        メールアドレスとパスワードをユーザーに求める。

        :param str | None mail: メールアドレス。
        :param str | None password: パスワード
        :rtype: dict[str, str]
        """
        un = mail
        pw = password
        try:
            if un is None:
                r = re.compile("^[a-zA-Z0-9.!#$%&'*+/=?^_`{|}~-]+@[a-zA-Z0-9-]+(?:\.[a-zA-Z0-9-]+)*$")
                while True:
                    print(Msg.input_mail)
                    un = input("-=>")
                    if not un: continue
                    if r.match(un): break
            if pw is None:
                while True:
                    print(Msg.input_pass)
                    pw = getpass("-=>")
                    if pw: break
        except (EOFError, KeyboardInterrupt):
            sys.exit(Err.keyboard_interrupt)
        return {
            "mail_tel": un,
            "password": pw
        }

    @classmethod
    def save_cookies(cls, requests_cookiejar, file_name=COOKIE_FILE_NAME):
        """
        クッキーを保存する。保存場所は基本的にユーザーのホームディレクトリ。

        :param requests.cookies.RequestsCookieJar requests_cookiejar:
        :param str file_name:
        :rtype: dict
        """
        # Python 3.5以上専用でいいならこう書く。
        # file_path = make_dir(Path.home() / file_name)
        # file_path.write_text("\n".join([k + "\t" + v for k, v in cook.items()]))
        file_path = join(expanduser("~"), file_name)
        cook = {key: val for key, val in requests_cookiejar.items()}
        with open(file_path, "w") as fd:
            fd.write("\n".join([k + "\t" + v for k, v in cook.items()]))
        return cook

    @classmethod
    def load_cookies(cls, file_name=COOKIE_FILE_NAME):
        """
        クッキーを読み込む。名前、値をタブで区切ったテキストファイルから。

        :param str file_name:
        :rtype: dict | None
        """
        # Python 3.5以上専用でいいならこう書く。
        # file_path = make_dir(Path.home() / file_name)
        #     return {line.split("\t")[0]: line.split("\t")[1]
        #             for line in file_path.read_text().split("\n")}
        try:
            file_path = join(expanduser("~"), file_name)
            with open(file_path, "r") as fd:
                return {line.split("\t")[0]: line.split("\t")[1].strip() for line in fd.readlines()}
        except (FileNotFoundError, EOFError):
            return None


class NTLogger(logging.Logger):
    def __init__(self, file_name=LOG_FILE_ND, name=__name__, log_level=logging.INFO):
        """
        ログ出力のためのクラス。

        :param str | Path | None file_name:
        :param str name:
        :param str | int log_level:
        """
        if not isinstance(log_level, (str, int)):
            raise ValueError("Invalid Logging Level: {}".format(log_level))

        log_level = logging.getLevelName(log_level)
        self.log_level = log_level
        if (IS_DEBUG or
            isinstance(log_level, str) and log_level == "DEBUG" or
            isinstance(log_level, int) and log_level <= logging.DEBUG):
            self._is_debug = True
        else:
            self._is_debug = False

        formatter = self.get_formatter()
        logging.Logger.__init__(self, name, log_level)
        self.logger = logging.getLogger(name=name)

        # 標準出力用ハンドラー
        log_stdout = logging.StreamHandler(sys.stdout)
        log_stdout.setLevel(log_level)
        log_stdout.setFormatter(formatter)
        self.addHandler(log_stdout)

        if file_name is not None:
            # ファイル書き込み用ハンドラー
            if isinstance(file_name, Path):
                log_file = logging.FileHandler(
                    filename=str(file_name), encoding="utf-8")
            else:
                log_file = logging.FileHandler(encoding="utf-8",
                    filename=join(expanduser("~"), file_name))
            log_file.setLevel(log_level)
            log_file.setFormatter(formatter)
            self.addHandler(log_file)

    def get_formatter(self):
        if self._is_debug:
            fmt = logging.Formatter("[{asctime}|{levelname: ^7}|{message}", style="{")
        else:
            fmt = logging.Formatter("[{asctime}|{levelname: ^7}]\t{message}", style="{")
        return fmt

    def forwarding(self, level, msg, *args, **kwargs):
        _enco = get_encoding()
        if self._is_debug:
            history = inspect.stack()
            funcs = "line:{}|{}".format(
                # ログを呼び出した場所の行数
                history[2][2],
                # ログを呼び出した関数の名前をつなげる
                " from ".join(["<{}>".format(item[3]) for item in history[2:5]])
            )
            if level <= logging.DEBUG:
                msg = funcs + "]\t" + str(msg)
            else:
                msg = funcs + "]\t\t" + str(msg)
        _msg = str(msg).encode(_enco, BACKSLASH).decode(_enco)
        _args = tuple([item.encode(_enco, BACKSLASH).decode(_enco)
                       if isinstance(item, str) else item for item in args[0]])
        self._log(level, _msg, _args, **kwargs)

    def debug(self, msg, *args, **kwargs): self.forwarding(logging.DEBUG, msg, args, **kwargs)

    def info(self, msg, *args, **kwargs): self.forwarding(logging.INFO, msg, args, **kwargs)

    def warning(self, msg, *args, **kwargs): self.forwarding(logging.WARNING, msg, args, **kwargs)

    def error(self, msg, *args, **kwargs): self.forwarding(logging.ERROR, msg, args, **kwargs)

    def critical(self, msg, *args, **kwargs): self.forwarding(logging.CRITICAL, msg, args, **kwargs)


class URL:
    URL_LogIn  = "https://secure.nicovideo.jp/secure/login?site=niconico"
    URL_Watch  = "http://www.nicovideo.jp/watch/"
    URL_GetFlv = "http://ext.nicovideo.jp/api/getflv/"
    URL_Info   = "http://ext.nicovideo.jp/api/getthumbinfo/"
    URL_Pict   = "http://tn-skr1.smilevideo.jp/smile"
    URL_GetThreadKey = "http://flapi.nicovideo.jp/api/getthreadkey"
    URL_WayBackKey = "http://flapi.nicovideo.jp/api/getwaybackkey"
    URL_Msg_JSON = "http://nmsg.nicovideo.jp/api.json/"
    URL_Msg_XML = "http://nmsg.nicovideo.jp/api/"

    # 一般のマイリストを扱うためのAPI
    URL_MyListTop  = "http://www.nicovideo.jp/my/mylist"
    URL_ListAll    = "http://www.nicovideo.jp/api/mylistgroup/list"
    URL_AddMyList  = "http://www.nicovideo.jp/api/mylistgroup/add"
    URL_PurgeList  = "http://www.nicovideo.jp/api/mylistgroup/delete"
    URL_ListOne    = "http://www.nicovideo.jp/api/mylist/list"
    URL_AddItem    = "http://www.nicovideo.jp/api/mylist/add"
    URL_DeleteItem = "http://www.nicovideo.jp/api/mylist/delete"
    URL_CopyItem   = "http://www.nicovideo.jp/api/mylist/copy"
    URL_MoveItem   = "http://www.nicovideo.jp/api/mylist/move"
    URL_UpdateItem = "http://www.nicovideo.jp/api/mylist/update"

    # とりあえずマイリストを扱うためのAPI
    URL_ListDef   = "http://www.nicovideo.jp/api/deflist/list"
    URL_AddDef    = "http://www.nicovideo.jp/api/deflist/add"
    URL_DeleteDef = "http://www.nicovideo.jp/api/deflist/delete"
    URL_CopyDef   = "http://www.nicovideo.jp/api/deflist/copy"
    URL_MoveDef   = "http://www.nicovideo.jp/api/deflist/move"
    URL_UpdateDef = "http://www.nicovideo.jp/api/deflist/update"


class Msg:
    """メッセージ集"""

    description = ("nicotools downlaod --help または nicotools mylist --help"
                   " で各コマンドのヘルプを表示します。")

    ''' マイリスト編集コマンドのヘルプメッセージ '''
    ml_description = ("マイリストを扱います。 add, delete, move, copy の引数には"
                      "テキストファイルも指定できます。 その場合はファイル名の"
                      "先頭に \"+\" をつけます。 例: +\"C:/ids.txt\"")
    ml_help_group_b = "マイリスト自体を操作する"
    ml_help_group_a = "リスト中の項目を操作する"
    ml_help_add = "指定したIDの動画を マイリストに追加します。"
    ml_help_delete = ("そのマイリストから 指定したIDの動画を削除します。"
                      "動画IDの代わりに * を指定すると、 マイリストを空にします。")
    ml_help_move = "移動元から移動先へと 動画を移動します。"
    ml_help_copy = ("コピー元からコピー先へと 動画をコピーします。 "
                    "動画IDの代わりに * を指定すると、 マイリスト全体をコピーします。")
    ml_help_export = ("登録された動画IDのみを改行で区切り、出力します。"
                      "名前の代わりに * を指定すると 全マイリストを一覧にします。")
    ml_help_show = ("登録された動画の情報をタブ区切り形式で出力します。"
                    "名前の代わりに * を指定すると マイリスト全体のメタデータを出力します。"
                    "-ss のように2回指定すると表形式で表示します。")
    ml_help_outfile = "そのファイル名で テキストファイルに出力します。"
    ml_help_purge = "そのマイリスト自体を削除します。 取り消しはできません。"
    ml_help_create = "指定した名前で 新しくマイリストを作成します。"
    ml_help_src = "移動(コピー)元、 あるいは各種の操作対象の、マイリストの名前"
    ml_help_to = "移動(コピー)先のマイリストの名前"
    ml_help_id = "マイリストの指定に、 名前の代わりにそのIDを使います。"
    ml_help_everything = ("show や export と同時に指定すると、"
                          "全てのマイリストの情報をまとめて取得します。")
    ml_help_yes = ("これを指定すると、マイリスト自体の削除や"
                   "マイリスト内の全項目の削除の時に確認しません。")
    ml_help_each = "指定すると、登録や削除を、まとめずに一つずつ行います。"

    ''' 動画ダウンロードコマンドのヘルプメッセージ '''
    nd_description = "動画のいろいろをダウンロードします。"
    nd_help_video_id = ("ダウンロードしたい動画ID。 例: sm12345678 "
                        "テキストファイルも指定できます。 その場合はファイル名の "
                        "先頭に \"+\" をつけます。 例: +\"C:/ids.txt\"")
    nd_help_password = "パスワード"
    nd_help_mail = "メールアドレス"
    nd_help_destination = "ダウンロードしたものを保存する フォルダーへのパス。"
    nd_help_outfile = "--getthumbinfo の結果をそのファイル名で テキストファイルに出力します。"
    nd_help_sieve = ("指定すると、動画とコメントについて、"
                     "非公開や削除済みの項目でもダウンロードを試みます。")
    nd_help_comment = "指定すると、 コメントをダウンロードします。"
    nd_help_video = "指定すると、 動画をダウンロードします。"
    nd_help_thumbnail = "指定すると、 サムネイルをダウンロードします。"
    nd_help_xml = ("指定すると、コメントをXML形式でダウンロードします。"
                   "チャンネル動画の場合は無視されます。")
    nd_help_info = "getthumbinfo API から動画の情報のみを ダウンロードします。"
    nd_help_what = "コマンドの確認用。 引数の内容を書き出すだけです。"
    nd_help_loglevel = "ログ出力の詳細さ。 デフォルトは INFO です。"
    nd_help_nomulti = "指定すると、プログレスバーを複数行で表示しません。"
    nd_help_limit = ("サムネイルとコメントについては同時ダウンロードを、"
                     "動画については1つあたりの分割数をこの数に制限します。標準は 4 です。")
    nd_help_dmc = "動画をDMCサーバー(いわゆる新サーバー)からダウンロードします。標準はこちらです。"
    nd_help_smile = "動画をsmileサーバー(いわゆる従来サーバー)からダウンロードします。"

    input_mail = "メールアドレスを入力してください。"
    input_pass = "パスワードを入力してください(画面には表示されません)。"

    ''' ログに書くメッセージ '''
    nd_start_download = "{count} 件の情報を取りに行きます。: {ids}"
    nd_download_done = "{path} に保存しました。"
    nd_download_video = "({0}/{1}) ID: {2} (タイトル:{3}) の動画をダウンロードします。"
    nd_download_pict = "({0}/{1}) ID: {2} (タイトル:{3}) のサムネイルをダウンロードします。"
    nd_download_comment = "({0}/{1}) ID: {2} (タイトル:{3}) のコメントをダウンロードします。"
    nd_start_dl_video = "{count} 件の動画をダウンロードします。: {ids}"
    nd_start_dl_pict = "{count} 件のサムネイルをダウンロードします。: {ids}"
    nd_start_dl_comment = "{count} 件のコメントをダウンロードします。: {ids}"
    nd_file_name = "{vid}_{name}.{ext}"
    nd_deleted_or_private = "{0} は削除されているか、非公開です。"

    ml_exported = "{0} に出力しました。"
    ml_items_counts = "含まれる項目の数:"
    ml_fetching_mylist_id = "マイリスト: {0} の ID を問い合わせています..."
    ml_showing_mylist = "マイリスト「{0}」の詳細を読み込んでいます..."
    ml_loading_mylists = "マイリストページを読み込んでいます..."
    ml_mylist_found = "ID: {0}, NAME: {1}, DESC: {2}"

    ml_ask_delete_all = "{0} に登録されている以下の全ての項目を削除します。"
    ml_confirmation = "この操作は取り消せません。よろしいですか? (Y/N)"
    ml_answer_yes = "処理を開始します。"
    ml_answer_no = "操作を中止しました。"
    ml_answer_invalid = "Y または N を入力してください。"
    ml_deleted_or_private = "{0[video_id]} {0[title]} は削除されているか非公開です。"

    ml_done_add = "[完了:追加] ({now}/{all}) 動画: {video_id}"
    ml_done_delete = "[完了:削除] ({now}/{all}) 動画: {video_id}"
    ml_done_copy = "[完了:コピー] ({now}/{all}) 動画ID: {video_id}"
    ml_done_move = "[完了:移動] ({now}/{all}) 動画ID: {video_id}"
    ml_done_purge = "[完了:マイリスト削除] 名前: {name}"
    ml_done_create = ("[完了:マイリスト作成] ID: {_id}, 名前:"
                      " {name} (公開: {pub}), 説明文: {desc}")

    ml_will_add = "[作業内容:追加] 対象: {0}, 動画ID: {1}"
    ml_will_delete = "[作業内容:削除] {0} から, 動画ID: {1}"
    ml_will_copy = "[作業内容:コピー] {0} から {1} へ, 動画ID: {2}"
    ml_will_move = "[作業内容:移動] {0} から {1} へ, 動画ID: {2}"
    ml_will_purge = "[作業内容:マイリスト削除] マイリスト「{0}」を完全に削除します。"


class Err:
    """ エラーメッセージ """

    failed_operation = "以下の理由により操作は失敗しました: {desc}"
    waiting_for_permission = "アクセス制限が解除されるのを待っています…"
    name_replaced = ("作成しようとした名前「{0}」は特殊文字を含むため、"
                     "「{1}」に置き換わっています。")
    cant_create = "この名前のマイリストは作成できません。"
    deflist_to_create_or_purge = "とりあえずマイリストは操作の対象にできません。"
    not_installed = "{0} がインストールされていないため実行できません。"
    invalid_argument = "引数の型が間違っています。"
    invalid_dirname = "このフォルダー名 {0} はシステム上使えません。他の名前を指定してください。"
    invalid_auth = "メールアドレスとパスワードを入力してください。"
    invalid_videoid = ("[エラー] 指定できる動画IDの形式は以下の通りです。"
                       "http://www.nicovideo.jp/watch/sm1234, "
                       "sm1234, nm1234, so1234,  123456, watch/123456")
    connection_404 = "404エラーです。 ID: {0} (タイトル: {1})"
    connection_timeout = "接続が時間切れになりました。 ID: {0}"
    keyboard_interrupt = "操作を中断しました。"
    not_specified = "[エラー] {0} を指定してください。"
    videoids_contain_all = "通常の動画IDと * を混ぜないでください。"
    list_names_are_same = "[エラー] 発信元と受信先の名前が同じです。"
    cant_perform_all = "[エラー] このコマンドに * は指定できません。"
    only_perform_all = "[エラー] このコマンドには * のみ指定できます。"
    unexpected_commands = "このコマンドは使用できません。 {0}"
    no_commands = "[エラー] コマンドを指定してください。"
    item_not_contained = "[エラー] 以下の項目は {0} に存在しません: {1}"
    name_ambiguous = ("同名のマイリストが {0}件あります。名前の代わりに"
                      "IDで(--id を使って)指定し直してください。")
    name_ambiguous_detail = ("ID: {id}, 名前: {name}, {publicity},"
                             " 作成日: {since}, 説明文: {description}")
    mylist_not_exist = "[エラー] {0} という名前のマイリストは存在しません。"
    mylist_id_not_exist = "[エラー] {0} というIDのマイリストは存在しません。"
    over_load = "[エラー] {0} にはこれ以上追加できません。"
    remaining = "以下の項目は処理されませんでした: {0}"
    already_exist = "[エラー]すでに存在しています。 ID: {0} (タイトル: {1})"
    known_error = "[エラー] 動画: {0}, コード: {}, 内容: {1}"
    unknown_error_itemid = "[エラー] ({0}/{1}) 動画: {2}, サーバーからの返事: {3}"
    unknown_error = "[エラー] ({0}/{1}) 動画: {2}, サーバーからの返事: {3}"
    failed_to_create = "[エラー] {0} の作成に失敗しました。 サーバーからの返事: {1}"
    failed_to_purge = "[エラー] {0} の削除に失敗しました。 サーバーからの返事: {1}"
    invalid_spec = ("[エラー] {0} は不正です。マイリストの名前"
                    "またはIDは文字列か整数で入力してください。")
    no_items = "[エラー] 指定した動画はいずれもこのマイリストには登録されていません。"

    '''
    APIから返ってくるエラーメッセージ
    {'error': {'code': 'MAXERROR', 'description': 'このマイリストにはもう登録できません'},'status':'fail'}
    {'error': {'code': 'MAXERROR', 'description': 'もうマイリストを作成できません', 'status': 'fail'}
    {'error': {'code': 'EXIST', 'description': 'すでに登録されています'}, 'status': 'fail'}
    {'error': {'code': 'NONEXIST', 'description': 'アイテムが存在しません'}, 'status': 'fail'}

    {'error': {'code': 'COMMANDERROR', 'description': '未定義のコマンドです'}, 'status': 'fail'}
    {'error': {'code': 'PARAMERROR', 'description': 'パラメータエラー'}, 'status': 'fail'}
    {'error': {'code': 'INVALIDTOKEN', 'description': '不正なトークンです'}, 'status': 'fail'}
    {'error': {'code': 'NOAUTH', 'description': '認証できませんでした'}, 'status': 'fail'}

    '''
    # エラーだが対処可能
    MAXERROR = "MAXERROR"
    EXIST = "EXIST"
    NONEXIST = "NONEXIST"

    # コマンドが不正
    COMMANDERROR = "COMMANDERROR"
    PARAMERROR = "PARAMERROR"
    INVALIDTOKEN = "INVALIDTOKEN"
    EXPIRETOKEN = "EXPIRETOKEN"
    NOAUTH = "NOAUTH"
    SECRETUSER = "SECRETUSER"
    INVALIDUSER = "INVALIDUSER"
    INVALIDVIDEO = "INVALIDVIDEO"

    # サーバーが不調
    MAINTENANCE = "MAINTENANCE"
    INTERNAL = "INTERNAL"


class KeyGTI:
    """
    getthumbinfo から取ってきたデータのキーとなる文字列たち。

    DATE ("first_retrieve")について:
        もともとは       2014-07-26T19:27:07+09:00
        という文字列だが 2014-07-26
        として保存しておく。

    MOVIE_TYPE ("movie_type")について:
        "mp4", "flv", "swf" のいずれか。

    URL_PIC ("thumbnail_url")について:
        例えば… http://tn-skr1.smilevideo.jp/smile?i=24093152
    """
    CH_ID           = "ch_id"           # チャンネル動画のみ
    CH_NAME         = "ch_name"         # チャンネル動画のみ
    CH_ICON_URL     = "ch_icon_url"     # チャンネル動画のみ
    COMMENT_NUM     = "comment_num"
    FIRST_RETRIEVE  = "first_retrieve"  # 例えば: 2014-07-26 もともとは: 2014-07-26T19:27:07+09:00
    DELETED         = "deleted"
    DESCRIPTION     = "description"
    EMBEDDABLE      = "embeddable"      # int; 0 or 1
    FILE_NAME       = "file_name"       # 元データには無い
    FILE_SIZE       = "file_size"       # 元データには無い
    LAST_RES_BODY   = "last_res_body"
    LENGTH          = "length"          # str
    LENGTH_SECONDS  = "length_seconds"  # int
    NUM_RES         = "num_res"         # int
    MYLIST_COUNTER  = "mylist_counter"  # int
    MOVIE_TYPE      = "movie_type"      # いずれか: mp4, flv, swf
    NO_LIVE_PLAY    = "no_live_play"    # int; 0 or 1
    SIZE_HIGH       = "size_high"       # int
    SIZE_LOW        = "size_low"        # int
    TAGS            = "tags"
    TAGS_LIST       = "tags_list"       # list
    THUMBNAIL_URL   = "thumbnail_url"   # 例えば: http://tn-skr1.smilevideo.jp/smile?i=24093152
    TITLE           = "title"
    USER_ID         = "user_id"         # int
    USER_NAME       = "user_nickname"
    USER_ICON_URL   = "user_icon_url"
    V_OR_T_ID       = "v_or_t_id"       # 普通は video_id と同一。so動画の時にのみ thread_id の値を入れる。
    VIDEO_ID        = "video_id"
    VIEW_COUNTER    = "view_counter"    # int
    WATCH_URL       = "watch_url"


class KeyDmc:
    FILE_NAME       = "file_name"
    FILE_SIZE       = "file_size"

    VIDEO_ID        = "video_id"
    VIDEO_URL_SM    = "video_url"       # Smile サーバーのほう
    TITLE           = "title"
    THUMBNAIL_URL   = "thumbnail_url"
    ECO             = "eco"             # int
    MOVIE_TYPE      = "movie_type"
    # IS_DMC          = "is_dmc"          # int or None
    DELETED         = "deleted"         # int
    IS_DELETED      = "is_deleted"      # bool
    IS_PUBLIC       = "is_public"       # bool
    IS_OFFICIAL     = "is_official"     # bool
    IS_PREMIUM      = "is_premium"      # bool
    USER_ID         = "user_id"         # int
    USER_KEY        = "user_key"
    MSG_SERVER      = "ms"
    THREAD_ID       = "thread_id"       # int
    THREAD_KEY      = "thread_key"

    API_URL         = "api_url"
    RECIPE_ID       = "recipe_id"
    CONTENT_ID      = "content_id"
    VIDEO_SRC_IDS   = "video_src_ids"   # list
    AUDIO_SRC_IDS   = "audio_src_ids"   # list
    HEARTBEAT       = "heartbeat"       # int
    TOKEN           = "token"
    SIGNATURE       = "signature"
    AUTH_TYPE       = "auth_type"
    C_K_TIMEOUT     = "content_key_timeout"     # int
    SVC_USER_ID     = "service_user_id"     # USER_ID とたぶん同じ
    PLAYER_ID       = "player_id"
    PRIORITY        = "priority"

    # ↓公式動画にだけある情報
    OPT_THREAD_ID   = "optional_thread_id"
    NEEDS_KEY       = "needs_key"


class KeyGetFlv:
    """ GetFLV を解釈するときのURLパラメーターのキー """
    THREAD_ID       = "thread_id"           # int
    LENGTH          = "l"                   # int
    VIDEO_URL       = "url"                 # str
    MSG_SERVER      = "ms"                  # str
    MSG_SUB         = "ms_sub"              # str
    USER_ID         = "user_id"             # int
    IS_PREMIUM      = "is_premium"          # int
    NICKNAME        = "nickname"            # str
    USER_KEY        = "userkey"             # str

    # ↓公式動画にだけある情報
    OPT_THREAD_ID   = "optional_thread_id"  # int
    NEEDS_KEY       = "needs_key"           # int


class MKey:
    """ マイリスト情報を読み取るときのJSONのキー """
    ID          = "id"
    NAME        = "name"
    IS_PUBLIC   = "is_public"
    PUBLICITY   = "publicity"
    SINCE       = "since"
    DESCRIPTION = "description"
    ITEM_DATA   = "item_data"


class InheritedParser(ArgumentParser):
    """ 文字コードで問題を起こさないために ArgumentParser を上書きするクラス """
    def _read_args_from_files(self, arg_strings):
        # expand arguments referencing files
        new_arg_strings = []
        for arg_string in arg_strings:

            # for regular arguments, just add them back into the list
            if not arg_string or arg_string[0] not in self.fromfile_prefix_chars:
                new_arg_strings.append(arg_string)

            # replace arguments referencing files with the file content
            else:
                try:
                    # ↓文字コードを指定しないとCP932で開いてしまいエラーになる
                    with open(arg_string[1:], encoding="utf-8") as args_file:
                        arg_strings = []
                        for arg_line in args_file.read().splitlines():
                            for arg in self.convert_arg_line_to_args(arg_line):
                                arg_strings.append(arg)
                        arg_strings = self._read_args_from_files(arg_strings)
                        new_arg_strings.extend(arg_strings)
                except OSError:
                    err = sys.exc_info()[1]
                    self.error(str(err))

        # return the modified argument list
        return new_arg_strings



from setuptools import setup

with open('README.rst') as f:
    readme = f.read()

setup(
    name='nicotools',
    version='1.0.0',
    packages=['nicotools'],
    # package_data={'nicotools': ['nicotools/stubs/*']},
    # include_package_data=True,
    url='https://github.com/mo-san/niconico-tools',
    license='MIT License',
    author='Masaki Taniguchi',
    author_email='window100@gmail.com',
    description=('Downloading videos, comments and thumbnails and'
                 ' handling your Mylists on niconico (ニコニコ動画).'),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: Japanese',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ],
    long_description=readme,
    keywords=['niconico', 'nicovideo', 'ニコニコ動画'],
    install_requires=[
        'requests',
        'prettytable',
        'bs4',
        'aiohttp',
        'tqdm'
    ],
    entry_points={
        'console_scripts': ['nicotools = nicotools.__init__:main']
    }
)
